# -*- coding: utf-8 -*-

import os
from pathlib import Path

import yaml

# This must be done BEFORE importing any other dask modules
from simmate.configuration import dask  # loads default cluster settings

from dask.distributed import LocalCluster
import dask_jobqueue

from prefect.agent.local import LocalAgent

from simmate.configuration.prefect.connect_to_dask import set_default_executor

ALLOWED_CLUSTER_TYPES = [
    "Local",
    "PBS",
    "SLURM",
    "SGE",
    "HTCondor",
    "LSF",
    "Moab",
    "OAR",
]

HEADER_ART = r"""
 ____            _       ____ _           _
|  _ \  __ _ ___| | __  / ___| |_   _ ___| |_ ___ _ __
| | | |/ _` / __| |/ / | |   | | | | / __| __/ _ \ '__|
| |_| | (_| \__ \   <  | |___| | |_| \__ \ ||  __/ |
|____/ \__,_|___/_|\_\  \____|_|\__,_|___/\__\___|_|

"""

# Load our default cluster type, agent name, and agent labels
def load_agent_settings():
    SIMMATE_DIRECTORY = os.path.join(Path.home(), "simmate")
    AGENT_YAML = os.path.join(SIMMATE_DIRECTORY, "prefect_agent.yaml")
    if os.path.exists(AGENT_YAML):
        with open(AGENT_YAML) as file:
            settings = yaml.full_load(file)
    else:
        settings = {}

    # update each entry if it's not set in the yaml file
    default_values = {
        "agent_name": None,
        "agent_labels": [],
        "dask_cluster_type": "Local",
    }
    for key, default_value in default_values.items():
        if not settings.get(key):
            settings[key] = default_value

    # make sure we have a legal cluster type
    if settings["dask_cluster_type"] not in ALLOWED_CLUSTER_TYPES:
        raise Exception(
            f"{settings['dask_cluster_type']} is not an allowed cluster type."
            f"Please choose one of the following: {ALLOWED_CLUSTER_TYPES}"
        )

    return settings


default_agent_settings = load_agent_settings()

# NOTE: if you are using a jobqueue cluster, this function assumes you have
# default values set in a Dask config file.
def run_cluster_and_agent(
    cluster_type=default_agent_settings["dask_cluster_type"],
    agent_name=default_agent_settings["agent_name"],
    agent_labels=default_agent_settings["agent_labels"],
    njobs=None,  # Required for jobqueue-based clusters
    # I include these three options because they are commonly overwritten when
    # setting up clusters. I need to revist this in the future though and consider
    # adding more or just accpeting a general **cluster_kwargs for input
    job_cpu=None,
    job_mem=None,
    walltime=None,
):

    ################# STEP 1: Configure our Dask Cluster. #################

    # check if the user wants to override default settings for the SLURM jobs.
    # Each of these will only be passed to Cluster() if they are set
    cluster_kwargs = {}
    if job_cpu:
        cluster_kwargs["job_cpu"] = job_cpu
    if job_mem:
        cluster_kwargs["job_mem"] = job_mem
    if walltime:
        cluster_kwargs["walltime"] = walltime

    # Start a local cluster with default settings if requested
    if cluster_type == "Local":
        cluster = LocalCluster(n_workers=njobs)

    # otherwise we have a jobqueue cluster
    else:
        # grab the class from the module. It will be named something like PBSCluster
        ClusterClass = getattr(dask_jobqueue, f"{cluster_type}Cluster")
        # Now initialize the cluster using the default Dask config.
        cluster = ClusterClass(**cluster_kwargs)
        if not njobs:
            raise Exception("You must provide nworkers when using a JobQueue cluster")
        cluster.scale(njobs)

    # print some info about the cluster for the user
    print(HEADER_ART)
    print(f"Scheduler is located at {cluster.scheduler.address}")
    print(f"Dashboard is located at {cluster.dashboard_link}")
    print(
        "To view the dashboard via ssh, use the command... \n\t"
        "ssh -N -L 8787:warwulf.net:8787 WarrenLab@warwulf.net"  # TODO: make more general
    )
    # Also if we have a jobqueue, we can display the submit script being used.
    if issubclass(cluster.__class__, dask_jobqueue.JobQueueCluster):
        print("Workers are submitted to SLURM with the following submit.sh...\n")
        print(cluster.job_script())
    else:
        print("Workers are all ran on the local machine.")
    print("\n\n")

    ################# STEP 2: Configure our Dask Cluster. #################

    # We want Prefect to use our Dask Cluster to run all of the workflow tasks. To
    # tell Prefect to do this, we wrote a helper function (set_default_executor) that
    # ships with Simmate.
    set_default_executor(cluster.scheduler.address)

    agent = LocalAgent(
        name=agent_name,
        labels=agent_labels,
        hostname_label=False,
    )

    # Now we can start the Prefect Agent which will run and search for jobs and then
    # submit them to our Dask cluster.
    agent.start()
    # NOTE: this line will run endlessly unless you set a timelimit in the LocalAgent above

    #######################################################################



# -*- coding: utf-8 -*-

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:

from simmate.calculators.vasp.workflows.band_structure import (
    workflow as matproj_workflow,
)



# -*- coding: utf-8 -*-

"""
This module defines the Structure class.

It is a very basic extension PyMatGen's core Structure class, as it only adds
a few extra methods and does not change any other usage.
"""

import os

from pymatgen.core import Structure as PymatgenStructure


class Structure(PymatgenStructure):
    # Leave docstring blank and just inherit from pymatgen

    def get_sanitized_structure(self):
        """
        Run symmetry analysis and "sanitization" on the pymatgen structure
        """

        # Make sure we have the primitive unitcell first
        # We choose to use SpagegroupAnalyzer (which uses spglib) rather than pymatgen's
        # built-in Structure.get_primitive_structure function:
        #   structure = structure.get_primitive_structure(0.1) # Default tol is 0.25

        # Default tol is 0.01, but we use a looser 0.1 Angstroms
        # from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
        # structure_primitive = SpacegroupAnalyzer(structure, 0.1).find_primitive()
        # BUG: with some COD structures, this symm analysis doesn't work. I need
        # to look into this more and figure out why.

        # Default tol is 0.25 Angstroms
        structure_primitive = self.get_primitive_structure()

        # Convert the structure to a "sanitized" version.
        # This includes...
        #   (i) an LLL reduction
        #   (ii) transforming all coords to within the unitcell
        #   (iii) sorting elements by electronegativity
        structure_sanitized = structure_primitive.copy(sanitize=True)

        # return back the sanitized structure
        return structure_sanitized

    @classmethod
    def from_dynamic(cls, structure):
        """
        This is an experimental feature.

        Possible structure formats include...
            object of toolkit structure
            dictionary of toolkit structure
            dictionary of...
                (1) python path to calculation datatable
                (2) one of the following (only one is used in this priority order):
                    (a) prefect flow id
                    (b) calculation id
                    (c) directory
                    ** these three are chosen because all three are unique for every
                    single calculation and we have access to different ones at different
                    times!
                (3) (optional) attribute to use on table (e.g. structure_final)
                    By default, we assume calculation table is also a structure table
            filename for a structure (cif, poscar, etc.) [TODO]
        """

        # keep track of this flag too, which we use in the next sections on loading
        # the directory and source
        is_from_past_calc = False

        # if the input is already a pymatgen structure, just return it back
        if isinstance(structure, PymatgenStructure):
            structure_cleaned = structure

        # if the "@module" key is in the dictionary, then we have a pymatgen
        # structure dict which we convert to a pymatgen object and return
        elif isinstance(structure, dict) and "@module" in structure.keys():
            structure_cleaned = cls.from_dict(structure)

        # if there is a calculation_table key, then we are pointing to the simmate
        # database for the input structure
        elif isinstance(structure, dict) and "calculation_table" in structure.keys():
            is_from_past_calc = True
            structure_cleaned = cls.from_database(structure)

        # if the value is a str and it relates to a filepath, then we load the
        # structure from a file.
        elif isinstance(structure, str) and os.path.exists(structure):
            structure_cleaned = cls.from_file(structure)

        # Otherwise an incorrect format was given
        else:
            raise Exception(
                "Unknown format provided for structure input. "
                f"{type(structure)} was provided. If you are trying "
                "to provide a filepath (str), make sure you don't have "
                "any typos and that the path is relative to the working "
                "directory."
            )

        # add this attribute to help with error checking in other methods
        structure_cleaned.is_from_past_calc = is_from_past_calc

        return structure_cleaned

    @classmethod
    def from_database(cls, structure: dict):
        """
        This is an experimental feature.

        Loads a structure from the Simmate database.
        """

        # because the structure is in the database, we need to setup django and
        # make sure we can access the tables. This import break the modularity
        # of this toolkit, so we therefore keep import within this function.
        from simmate.configuration.django import setup_full  # connects to database
        from simmate.website.workflows import models as all_datatables
        from django.utils.module_loading import import_string

        # start by loading the datbase table, which is given as a module path
        datatable_str = structure["calculation_table"]

        # Import the datatable class -- how this is done depends on if it
        # is from a simmate supplied class or if the user supplied a full
        # path to the class
        # OPTIMIZE: is there a better way to do this?
        if hasattr(all_datatables, datatable_str):
            datatable = getattr(all_datatables, datatable_str)
        else:
            datatable = import_string(datatable_str)

        # These attributes tells us which structure to grab from our datatable.
        # The user should have only provided one -- if they gave more, we just
        # use whichever one comes first.
        prefect_flow_run_id = structure.get("prefect_flow_run_id")
        calculation_id = structure.get("calculation_id")
        directory_old = structure.get("directory")

        # we must have either a prefect_flow_run_id or calculation_id
        if not prefect_flow_run_id and not calculation_id and not directory_old:
            raise Exception(
                "You must have either a prefect_flow_run_id, calculation_id, "
                "or directory provided if you want to load a structure from "
                "a previous calculation."
            )

        # now query the datable with which whichever was provided. Each of these
        # are unique so all three should return a single calculation.
        if calculation_id:
            calculation = datatable.objects.get(id=calculation_id)
        elif prefect_flow_run_id:
            calculation = datatable.objects.get(
                prefect_flow_run_id=prefect_flow_run_id,
            )
        elif directory_old:
            calculation = datatable.objects.get(directory=directory_old)

        # In some cases, the structure we want is not within the calculation table.
        # For example, in relaxations the final structure is attached via
        # the table.structure_final attribute
        structure_field = structure.get("structure_field")
        if structure_field:
            structure_cleaned = getattr(calculation, structure_field).to_toolkit()
        # if there's no structure field, that means we already have the correct entry
        else:
            structure_cleaned = calculation.to_toolkit()

        # For ease of access, we also link the calculation database entry
        structure_cleaned.calculation = calculation

        return structure_cleaned

    @classmethod
    def from_database_string(cls, structure_string: str):
        """
        Loads a toolkit structure from a string -- specifically strings that
        are stored in the structure_string column for
        simmate.database.base_data_types.Structure.
        """
        # Dev note: this method should be merged with the Toolkit.from_str method.
        # I only have this separate for now because pymatgen's from_str doesn't
        # dynamically determine format from the string alone.

        # convert the stored string to python dictionary.
        storage_format = "CIF" if (structure_string[0] == "#") else "POSCAR"
        # OPTIMIZE: see my comment on storing strings in the from_toolkit method above.
        # For now, I need to figure out if I used "CIF" or "POSCAR" and read the structure
        # accordingly. In the future, I can just assume my new format.
        # If the string starts with "#", then I know that I stored it as a "CIF".

        # convert the string to pymatgen Structure object
        if storage_format == "POSCAR":
            structure = cls.from_str(
                structure_string,
                fmt=storage_format,
            )

        # BUG: for cod structures we need to set the tolerance to "inf", which
        # isn't possible with the toolkit.from_str method. Instead we need to
        # call the CifParser directly
        elif storage_format == "CIF":
            from pymatgen.io.cif import CifParser

            parser = CifParser.from_string(
                structure_string,
                occupancy_tolerance=float("inf"),
            )
            structure = parser.get_structures()[0]

        return structure



# -*- coding: utf-8 -*-

from django.shortcuts import render

from simmate.database import third_parties
from simmate.website.core_components.utilities import render_from_table


def providers_all(request):

    # TODO: auto determine this list and descriptions
    workflows_metadata = {
        # "AflowStructure": (
        #     "These workflows calculate the energy for a structure. In many cases, "
        #     "this also involves calculating the lattice strain and forces for each site."
        # ),
        "CodStructure": "...",
        "JarvisStructure": "...",
        "MatProjStructure": "...",
        "OqmdStructure": "...",
    }

    # now let's put the data and template together to send the user
    context = {"workflows_metadata": workflows_metadata}
    template = "third_parties/providers_all.html"
    return render(request, template, context)


def provider(request, provider_name: str):

    # using the provider name (which is really just the table name), load
    # the corresponding database table
    provider_table = getattr(third_parties, provider_name)

    return render_from_table(
        request=request,
        template="third_parties/provider.html",
        context={"provider": provider_table},
        table=provider_table,
        view_type="list",
    )


def entry_detail(
    request,
    provider_name: str,
    entry_id: int,
):

    # using the provider name (which is really just the table name), load
    # the corresponding database table
    provider_table = getattr(third_parties, provider_name)

    return render_from_table(
        request=request,
        request_kwargs={
            "provider_name": provider_name,
            "entry_id": entry_id,
        },
        template="third_parties/entry_detail.html",
        context={},
        table=provider_table,
        view_type="retrieve",
        primary_key_url="entry_id",
    )



# -*- coding: utf-8 -*-

from .quality_00 import Quality00Relaxation
from .quality_01 import Quality01Relaxation
from .quality_02 import Quality02Relaxation
from .quality_03 import Quality03Relaxation
from .quality_04 import Quality04Relaxation
from .mit import MITRelaxation
from .materials_project import MatProjRelaxation
from .neb_endpoint import NEBEndpointRelaxation



# -*- coding: utf-8 -*-

# NOTE: This task isn't used at all yet because my update_calculation method
# of NestedCalculation is broken.

import prefect
from prefect import Task


class SaveNestedCalculationTask(Task):
    def __init__(self, calculation_table, **kwargs):
        self.calculation_table = calculation_table
        super().__init__(**kwargs)

    def run(self):

        # first grab the calculation entry
        calc = self.calculation_table.from_prefect_id(prefect.context.flow_run_id)

        # now update the calculation
        calc.update_calculation()



# -*- coding: utf-8 -*-

import os
import itertools
import sys


def get_conda_env() -> str:
    """
    Grab the name of the activate conda environment and returns it as a string.
    If there is no env, then an empty string is returned.
    """
    # Check the list of python paths and grab the first path that has "envs" in it.
    # Assume we don't have a conda env until proven otherwise
    env_name = ""
    for path in sys.path:
        if "envs" in path:
            # split the path into individual folder names (os.sep gives / or \\)
            folders = path.split(os.sep)
            # the conda env name will the name immediately after the /envs.
            # example path is '/home/jacksund/anaconda3/envs/simmate_dev/lib/python3.10'
            # where we want the name simmate_dev here.
            env_name = folders[folders.index("envs") + 1]
            # once we have found this, we can exit the loop
            break

    return env_name


def get_doc_from_readme(file: str) -> str:
    """
    Loads the docstring from a README.md file in the same directory.

    This is commonly used in __init__.py files because we like having our
    documentation isolated (so that github renders it).

    To use, simply pass the file property:

    ``` python
    from simmate.utilities import get_doc_from_readme

    __doc__ = get_doc_from_readme(__file__)
    ```

    This is an alternative to using "include" in rst files, which
    [pdoc recommends](https://pdoc.dev/docs/pdoc.html#include-markdown-files).
    We prefer this utility because it allows Spyder to load the docs -- although
    it's slower in production (bc of opening/closing files).
    """

    # We assume the file is in the same directory and named "README.rst"
    file_directory = os.path.dirname(os.path.abspath(file))
    with open(
        os.path.join(file_directory, "README.md"),
        encoding="utf-8",
    ) as doc_file:
        doc = doc_file.read()
    return doc


def get_chemical_subsystems(chemical_system: str):
    """
    Given a chemical system, this returns all chemical systems that are also
    contained within it.

    For example, "Y-C" would return ["Y", "C", "C-Y"]. Note that the returned
    list has elements of a given system in alphabetical order (i.e. it gives
    "C-Y" and not "Y-C")

    #### Parameters

    - `chemical_system`:
        A chemical system of elements. Elements must be separated by dashes (-)

    #### Returns

    - `subsystems`:
        A list of chemical systems that make up the input chemical system.
    """

    # TODO: this will may be better located elsewhere. Maybe even as a method for
    # the Composition class.

    # Convert the system to a list of elements
    system_cleaned = chemical_system.split("-")

    # Now generate all unique combinations of these elements. Because we also
    # want combinations of different sizes (nelements = 1, 2, ... N), then we
    # put this in a for-loop.
    subsystems = []
    for i in range(len(system_cleaned)):
        # i is the size of combination we want. We now ask for each unique combo
        # of elements at this given size.
        for combo in itertools.combinations(system_cleaned, i + 1):
            # Combo will be a tuple of elements that we then convert back to a
            # chemical system. We also sort this alphabetically.
            #   ex: ("Y", "C", "F") ---> "C-F-Y"
            subsystem = "-".join(sorted(combo))
            subsystems.append(subsystem)
    return subsystems



# -*- coding: utf-8 -*-

# django has builtin views for login/logout/etc
# import if you want to setup login/logout views with more control
# from django.contrib.auth import views as auth_views

from django.urls import include, path

from simmate.website.accounts import views

urlpatterns = [
    # This points to a built-in app for django that has the some login/password options
    # https://docs.djangoproject.com/en/3.0/topics/auth/default/#module-django.contrib.auth.views
    path(route="", view=include("django.contrib.auth.urls")),
    # the include() here maps the following:
    #   accounts/login/ [name='login']
    #   accounts/logout/ [name='logout']
    #   accounts/password_change/ [name='password_change']
    #   accounts/password_change/done/ [name='password_change_done']
    #   accounts/password_reset/ [name='password_reset']
    #   accounts/password_reset/done/ [name='password_reset_done']
    #   accounts/reset/<uidb64>/<token>/ [name='password_reset_confirm']
    #   accounts/reset/done/ [name='password_reset_complete']
    # The built-in app doesn't provide a way to create new users though
    path(
        route="register/",
        view=views.register,
        name="register",
    ),
    # On login success, you will be pointed to /accounts/profile by default,
    # but this isn't mapped. If you want to change this defualt, then
    # set LOGIN_REDIRECT_URL in the settings.py file
    path(
        route="profile/",
        view=views.profile,
        name="profile",
    ),
    # When you point to the url 'logout', it signs you out and then sends you
    # to LOGOUT_REDIRECT_URL (set in settings.py)
    path(
        route="loginstatus/",
        view=views.loginstatus,
        name="loginstatus",
    ),
]



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.nudged_elastic_band import (
    all_paths_workflow,
    single_path_workflow,
    from_endpoints_workflow,
    from_images_workflow,
)



from django.shortcuts import render, redirect

from django.contrib.auth import authenticate, login
from django.contrib.auth.decorators import login_required
from django.contrib.auth.forms import UserCreationForm


def register(request):

    if request.method == "POST":
        usercreation_form = UserCreationForm(request.POST)
        if usercreation_form.is_valid():
            # create the user
            new_user = usercreation_form.save()
            # now automatically sign them in
            new_user = authenticate(
                username=usercreation_form.cleaned_data["username"],
                password=usercreation_form.cleaned_data["password1"],
            )
            login(request, new_user)
            # after first creating a profile, I send them to their profile page
            return redirect("profile")
    else:
        usercreation_form = UserCreationForm()

    context = {"usercreation_form": usercreation_form}
    template = "registration/register.html"
    return render(request, template, context)


@login_required
def profile(request):
    # !!! For future reference, you can grab user-associated data via...
    # data = request.user.relateddata.all()

    context = {}
    template = "registration/profile.html"
    return render(request, template, context)


def loginstatus(request):
    context = {}
    template = "registration/loginstatus.html"
    return render(request, template, context)


"""
The following views are mapped automatically, so for reference I write out
the template names below. To see the source code, visit
https://github.com/django/django/blob/master/django/contrib/auth/views.py

    accounts/login/ [name='login']
        template = 'registration/login.html' by default
        on successful login, redirects to settings.LOGIN_REDIRECT_URL

    accounts/logout/ [name='logout']
        template = 'registration/logged_out.html' by default
        on successful logout, redirects to settings.LOGOUT_REDIRECT_URL



    # TODO: the below views are not fully setup yet

    accounts/password_change/ [name='password_change']
        template_name = 'registration/password_change_form.html'

    accounts/password_change/done/ [name='password_change_done']
        template_name = 'registration/password_change_done.html'

    accounts/password_reset/ [name='password_reset']
        email_template = 'registration/password_reset_email.html'
        subject_template = 'registration/password_reset_subject.txt'
        template = 'registration/password_reset_form.html'

    accounts/password_reset/done/ [name='password_reset_done']
        template = 'registration/password_reset_done.html'

    accounts/reset/<uidb64>/<token>/ [name='password_reset_confirm']
        template = 'registration/password_reset_confirm.html'

    accounts/reset/done/ [name='password_reset_complete']
        template_name = 'registration/password_reset_complete.html'
"""



# -*- coding: utf-8 -*-

# !!! Should this module be moved to simmate.workflows.common_tasks? There's
# nothing unique to VASP here

import os

from simmate.toolkit import Structure
from simmate.toolkit.diffusion import (
    DistinctPathFinder,
    MigrationHop,
    MigrationImages,
)

from simmate.database.base_data_types.nudged_elastic_band import (
    DiffusionAnalysis as DiffusionAnalysisTable,
    MigrationHop as MigrationHopTable,
    MigrationImage as MigrationImageTable,
)

from simmate.workflow_engine.workflow import Task, task

from typing import List, Tuple


class BuildDiffusionAnalysisTask(Task):
    def __init__(self, diffusion_analyis: DiffusionAnalysisTable, **kwargs):
        self.diffusion_analyis = diffusion_analyis
        super().__init__(**kwargs)

    def run(
        self,
        structure: Structure,
        migrating_specie: str,
        vacancy_mode: bool,
        directory: str = "",
        **kwargs,
    ) -> List[MigrationHop]:
        """
        Given a bulk crystal structure, returns all symmetrically unique pathways
        for the migrating specie (up until the path is percolating). This
        also create all relevent database entries for this struture and its
        migration hops.

        #### Parameters

        - `structure`:
            bulk crystal structure to be analyzed. Can be in any format supported
            by Structure.from_dynamic method.

        - `migrating_specie`:
            Element or ion symbol of the diffusion specie (e.g. "Li")

        - `directory`:
            where to write the CIF file visualizing all migration hops. If no
            directory is provided, it will be written in the working directory.

        - `**kwargs`:
            Any parameter normally accepted by DistinctPathFinder
        """

        ###### STEP 1: creating the toolkit objects and writing them to file

        structure_cleaned = Structure.from_dynamic(structure)

        pathfinder = DistinctPathFinder(
            structure_cleaned,
            migrating_specie,
            **kwargs,
        )
        migration_hops = pathfinder.get_paths()

        # We write all the path files so users can visualized them if needed
        filename = os.path.join(directory, "migration_hop_all.cif")
        pathfinder.write_all_paths(filename, nimages=10)
        for i, migration_hop in enumerate(migration_hops):
            number = str(i).zfill(2)  # converts numbers like 2 to "02"
            # the files names here will be like "migration_hop_02.cif"
            migration_hop.write_path(
                os.path.join(directory, f"migration_hop_{number}.cif"),
                nimages=10,
            )

        ###### STEP 2: creating the database objects and saving them to the db

        # Create the main DiffusionAnalysis object that others will link to.
        da_obj = self.diffusion_analyis.from_toolkit(
            structure=structure_cleaned,
            migrating_specie=migrating_specie,
            vacancy_mode=vacancy_mode,
        )
        da_obj.save()
        # TODO: should I search for a matching bulk structure before deciding
        # to create a new DiffusionAnalysis entry?

        # grab the linked MigrationHop class
        hop_class = da_obj.migration_hops.model

        # Now iterate through the hops and add them to the database
        hop_ids = []
        for i, hop in enumerate(migration_hops):
            hop_db = hop_class.from_toolkit(
                migration_hop=hop,
                number=i,
                diffusion_analysis_id=da_obj.id,
            )
            hop_db.save()
            hop_ids.append(hop_db.id)

        # TODO: still figuring out if toolkit vs. db objects should be returned.
        # Maybe add ids to the toolkit objects? Or dynamic DB dictionaries?
        # For now I return the MigrationHop ids -- because this let's me
        # indicate which MigrationHops should be updated later on.
        return hop_ids


@task(nout=2)
def get_endpoint_structures(migration_hop: MigrationHop) -> Tuple[Structure]:
    """
    Simple wrapper for get_sc_structures method that makes it a Prefect task.
    I assume parameters for now
    """
    start_supercell, end_supercell, _ = migration_hop.get_sc_structures(
        vac_mode=True,
    )
    try:
        assert start_supercell != end_supercell
    except:
        raise Exception(
            "This structure has a bug due to a rounding error. "
            "Our team is aware of this bug and it has been fixed for the next "
            "pymatgen-analysis-diffusion release."
        )
    return start_supercell, end_supercell


@task
def get_migration_images_from_endpoints(supercell_start, supercell_end):
    """
    Simple wrapper for from_endpoints method that makes it a Prefect task.
    I assume parameters for now.
    """

    # Make sure we have toolkit objects, and if not, convert them
    supercell_start_cleaned = Structure.from_dynamic(supercell_start)
    supercell_end_cleaned = Structure.from_dynamic(supercell_end)

    images = MigrationImages.from_endpoints(
        structure_start=supercell_start_cleaned,
        structure_end=supercell_end_cleaned,
        nimages=7,  # TODO: have from_endpoints figure out pathway length
    )

    return images


class SaveNEBOutputTask(Task):
    # This is a modification of simmate.workflows.common_tasks.SaveOutputTask

    def __init__(
        self,
        diffusion_analyis_table: DiffusionAnalysisTable,
        migration_hop_table: MigrationHopTable,
        migration_image_table: MigrationImageTable,
        **kwargs,
    ):
        self.diffusion_analyis_table = diffusion_analyis_table
        self.migration_hop_table = migration_hop_table
        self.migration_image_table = migration_image_table
        super().__init__(**kwargs)

    def run(
        self,
        output,
        diffusion_analysis_id: int = None,
        migration_hop_id: int = None,
    ):

        # split our results and corrections (which are given as a dict) into
        # separate variables
        # Our result here is not a VaspRun object, but instead a NEBAnalysis
        # object. See NudgedElasticBandTask.workup()
        result = output["result"]

        # TODO: These aren't saved for now. Consider making MigrationHopTable
        # a Calculation and attaching these there.
        corrections = output["corrections"]
        directory = output["directory"]

        # First, we need a migration_hop database object.
        # All of hops should link to a diffusion_analysis entry, so we check
        # for that here too. The key thing of these statements is that we
        # have a migration_hop_id at the end.

        migration_hop_db = self.migration_hop_table.from_pymatgen(
            analysis=result,
            diffusion_analysis_id=diffusion_analysis_id,
            migration_hop_id=migration_hop_id,
        )

        # If the user wants to access results, they can do so through the hop id
        return migration_hop_db.id



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation.mit import MITRelaxation


class NEBEndpointRelaxation(MITRelaxation):
    """
    Runs a VASP relaxation calculation using MIT Project settings, where some
    settings are adjusted to accomodate large supercells with defects. Most
    notably, the lattice remains fixed and symmetry is turned off for this
    relaxation.

    These settings are closely related to relaxation/mit, but only meant to be
    used on start/end supercell structures of a NEB calculation.

    You typically shouldn't use this workflow directly, but instead use the
    higher-level NEB workflows (e.g. diffusion/neb_all_paths or
    diffusion/neb_from_endpoints), which call this workflow for you.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    # These settings are based off of pymatgen's MVLCINEBEndPointSet
    # http://guide.materialsvirtuallab.org/pymatgen-analysis-diffusion/pymatgen.analysis.diffusion.neb.io.html
    incar = MITRelaxation.incar.copy()
    incar.update(
        dict(
            ISIF=2,  # hold lattice volume and shape constant
            EDIFFG=-0.02,
            ISMEAR=0,
            ISYM=0,
            LDAU=False,
            NELMIN=4,
        )
    )
    # We set ISMEAR=0 and SIGMA above, so we no longer need smart_ismear
    incar.pop("multiple_keywords__smart_ismear")
    # LDA+U is turned off
    incar.pop("multiple_keywords__smart_ldau")



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from simmate.calculators.vasp.database.relaxation import (
    MatProjRelaxation,
    MatProjIonicStep,
    MITRelaxation,
    MITIonicStep,
    Quality00Relaxation,
    Quality00IonicStep,
    Quality01Relaxation,
    Quality01IonicStep,
    Quality02Relaxation,
    Quality02IonicStep,
    Quality03Relaxation,
    Quality03IonicStep,
    Quality04Relaxation,
    Quality04IonicStep,
    StagedRelaxation,
    NEBEndpointRelaxation,
    NEBEndpointIonicStep,
)

from simmate.calculators.vasp.database.energy import (
    MatProjStaticEnergy,
    MITStaticEnergy,
    Quality04StaticEnergy,
    NEBEndpointStaticEnergy,
)

from simmate.calculators.vasp.database.band_structure import (
    MatProjBandStructure,
)

from simmate.calculators.vasp.database.density_of_states import (
    MatProjDensityOfStates,
)

from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
    MITMigrationHop,
    MITMigrationImage,
)

from simmate.calculators.vasp.database.dynamics import (
    MITDynamicsRun,
    MITDynamicsIonicStep,
)

# This module is still under early development
# from simmate.toolkit.structure_prediction.evolution.database import (
#     EvolutionarySearch,
#     StructureSource,
# )



# -*- coding: utf-8 -*-

# NOTE TO DEVS: This workflow is very close to those made by s3task_to_workflow
# but there are some key differences, which is why we manually make the workflow
# here. Most notably, there is no "structures" parameter here -- instead, this
# flow takes migration_images. The fact that we are passing multiple structures
# changes the naming of some variables in the workflow. For example, we use
# neb_task(structures=...) instead of s3task(structure=...). Further, we also
# accept extra parameters that help link to other tables -- for high-level
# NEB workflows (where many images/hops are linked together). These parameters
# are used in saving the results.

from simmate.toolkit.diffusion import MigrationImages

from simmate.workflow_engine.workflow import (
    Workflow,
    Parameter,
    ModuleStorage,
)
from simmate.workflow_engine.common_tasks import LoadInputAndRegister
from simmate.calculators.vasp.tasks.nudged_elastic_band import MITNudgedElasticBand
from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
    MITMigrationHop,
    MITMigrationImage,
)
from simmate.calculators.vasp.workflows.nudged_elastic_band.utilities import (
    SaveNEBOutputTask,
)

WORKFLOW_NAME = "diffusion/from-images"

neb_task = MITNudgedElasticBand()
load_input_and_register = LoadInputAndRegister(
    workflow_name=WORKFLOW_NAME,
    input_obj_name="migration_images",
)
save_results = SaveNEBOutputTask(
    MITDiffusionAnalysis,
    MITMigrationHop,
    MITMigrationImage,
)

with Workflow(WORKFLOW_NAME) as workflow:

    migration_images = Parameter("migration_images")
    command = Parameter("command", default="vasp_std > vasp.out")
    source = Parameter("source", default=None)
    directory = Parameter("directory", default=None)
    # assume use_previous_directory=False for now

    # These parameters link to higher-level tables.
    diffusion_analysis_id = Parameter("diffusion_analysis_id", default=None)
    migration_hop_id = Parameter("migration_hop_id", default=None)

    # Load our input and make a base directory for all other workflows to run
    # within for us.
    migration_images_toolkit, directory_cleaned = load_input_and_register(
        input_obj=migration_images,
        input_class=MigrationImages,
        source=source,
        directory=directory,
    )

    output = neb_task(
        structures=migration_images_toolkit,
        command=command,
        directory=directory_cleaned,
    )

    calculation_id = save_results(
        output=output,
        diffusion_analysis_id=diffusion_analysis_id,
        migration_hop_id=migration_hop_id,
    )  # calculation_id will correspond to the migration hop table


workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-Diffusion"
workflow.calculation_table = MITMigrationImage
workflow.result_table = MITMigrationImage
workflow.register_kwargs = ["prefect_flow_run_id"]
workflow.result_task = output
workflow.s3task = neb_task

# workflow.calculation_table = MITDiffusionAnalysis  # not implemented yet
# workflow.register_kwargs = ["prefect_flow_run_id"]

# workflow.__doc__ = neb_task.__doc__
workflow.__doc__ = """
    Runs a NEB relaxation on a list of structures (aka images) using MIT Project
    settings. The lattice remains fixed and symmetry is turned off for this
    relaxation.
    
    Typically, this workflow is not ran directly -- but instead, you would
    use diffusion/all-paths which submits a series of this workflow for you. Other
    higher level workflows that call this one include diffusion/from-endpoints
    and diffusion/single-path.
"""



# --------------------------------------------------------------------------------------

# import pickle
import cloudpickle  # needed to serialize Prefect workflow runs and tasks

from simmate.database.base_data_types import DatabaseTable, table_column

# TYPES OF RELATIONSHIPS:
# ManyToMany - place in either but not both
# ManyToOne (ForeignKey) - place in the many
# OneToOne - place in the one that has extra features (it's like setting a parent class)

# --------------------------------------------------------------------------------------

# For serialization, I just use the pickle module, but in the future, I may
# want to do a priority of MsgPk >> JSON >> Pickle. I could check if the given
# object(s) has a serialize() method, check if has a to_json() method, and then
# as a last resort just pickle the object.
# https://docs.python.org/3/library/pickle.html

# Pickled objects are just written as byte strings, so I stored them in django
# as a BinaryField which accepts bytes.
# https://docs.djangoproject.com/en/3.1/ref/models/fields/#binaryfield

# There is a repo that goes one step further and makes a PickleField for django
# but this looks over the top for what I need -- even though django-picklefield
# is only one file! Still, I can use it as a reference if I ever run into bugs
# https://github.com/gintas/django-picklefield

# The name WorkItem comes from the naming convention use here:
# https://github.com/python/cpython/blob/master/Lib/concurrent/futures/thread.py

# --------------------------------------------------------------------------------------


class WorkItem(DatabaseTable):

    """Base info"""

    # The function to be called
    fxn = table_column.BinaryField()

    # arguments to be passed into fxn
    args = table_column.BinaryField(default=cloudpickle.dumps([]))

    # keyword arguments to be passed into fxn
    kwargs = table_column.BinaryField(default=cloudpickle.dumps({}))

    # the output of fxn(*args, **kwargs)
    result = table_column.BinaryField(blank=True, null=True)

    # the status/state of the workitem
    # These states are based on the python queue module
    class StatusOptions(table_column.TextChoices):
        PENDING = "P"
        RUNNING = "R"
        CANCELLED = "C"
        # CANCELLED_AND_NOTIFIED = "N"  # !!! when should I use this?
        FINISHED = "F"

    # TODO -- I should consider indexing this column for speed because it's
    # the most queried column by far.
    status = table_column.CharField(
        max_length=1,
        choices=StatusOptions.choices,
        default=StatusOptions.PENDING,
    )

    # TODO -- This really should be a separate table with a relationship to WorkItem
    # the worker ID that grabbed the workitem
    # worker_id = table_column.CharField(max_length=50, blank=True, null=True)

    class Meta:
        app_label = "workflow_execution"



# -*- coding: utf-8 -*-

import prefect
from prefect import Task


class SaveOutputTask(Task):
    def __init__(self, calculation_table, **kwargs):
        self.calculation_table = calculation_table
        super().__init__(**kwargs)

    def run(self, output):

        # split our results and corrections (which are given as a dict) into
        # separate variables
        vasprun = output["result"]
        corrections = output["corrections"]
        directory = output["directory"]

        # load the calculation entry for this workflow run. This should already
        # exist thanks to the load_input_and_register task.
        calculation = self.calculation_table.from_prefect_id(
            prefect.context.flow_run_id,
        )

        # now update the calculation entry with our results
        calculation.update_from_vasp_run(vasprun, corrections, directory)

        return calculation.id



# -*- coding: utf-8 -*-

from .atomic_permutation import AtomicPermutation
from .atomic_permutation_ase import AtomicPermutationASE
from .coordinate_perturation_ordered import CoordinateOrderedPerturbation
from .coordinate_perturbation_ase import CoordinatePerturbationASE
from .coordinate_perturbation import CoordinatePerturbation
from .heredity_mutation_ase import HeredityASE
from .lattice_strain_ase import LatticeStrainASE
from .lattice_strain import LatticeStrain
from .mirror_mutation_ase import MirrorMutationASE
from .rotational_mutation_ase import RotationalMutationASE
from .soft_mutation_ase import SoftMutationASE



# -*- coding: utf-8 -*-

from simmate.workflow_engine import s3task_to_workflow
from simmate.calculators.vasp.tasks.band_structure import (
    MatProjBandStructure as MatProjBandStructureTask,
)
from simmate.calculators.vasp.database.band_structure import (
    MatProjBandStructure as MatProjBandStructureResults,
)

workflow = s3task_to_workflow(
    name="band-structure/matproj",
    module=__name__,
    project_name="Simmate-Band-Structure",
    s3task=MatProjBandStructureTask,
    calculation_table=MatProjBandStructureResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)



# -*- coding: utf-8 -*-

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:
from simmate.calculators.vasp.workflows.energy import (
    mit_workflow,
    quality04_workflow,
    matproj_workflow,
    neb_endpoint_workflow,
)



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.electronic_structure.plotter import BSPlotter

from simmate.calculators.vasp.inputs import Incar, Poscar, Potcar
from simmate.calculators.vasp.tasks.energy.materials_project import (
    MatProjStaticEnergy,
)


class MatProjBandStructure(MatProjStaticEnergy):
    """
    Calculates the band structure using Materials Project settings.

    Your stucture will be converted to the standardized-primitive unitcell so
    that the high-symmetry K-point path can be used.

    This is also a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which handles the preliminary relaxation and SCF
    energy calculation for you. This S3Task is only the final step of that workflow.

    See `vasp.workflows.band_structure`.
    """

    # The default settings to use for this calculation.
    # The key thing for band structures is that this follows a static energy
    # calculation. We use the CHGCAR from that previous calculation. We also
    # hold the charge density static during this calculation so this is a
    # non-selfconsistent (non-SCF) calculation.
    incar = MatProjStaticEnergy.incar.copy()
    incar.update(
        ICHARGE=11,
        ISMEAR=0,
        SIGMA=0.01,
    )

    # set the KptGrid or KptPath object
    # TODO: in the future, all functionality of this class will be available
    # by giving a KptPath class here.
    kpoints = None
    kpoints_line_density = 20

    # For band-structures, unit cells should be in the standardized format
    pre_standardize_structure = True

    def setup(self, structure, directory):
        """
        Writes input files for this calculation. This differs from the normal
        VaspTask setup because it converts the structure to the standard primative
        first and then writes a KPOINT file with using a highsym path.
        """

        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = self._get_clean_structure(structure)

        # write the poscar file
        Poscar.to_file(structure_cleaned, os.path.join(directory, "POSCAR"))

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**self.incar) + Incar(**self.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            structure=structure_cleaned,
        )

        ##############
        # we need to find the high-symmetry Kpt path. Note that all of this
        # functionality will be moved to the KptPath class and then extended to
        # vasp.inputs.kpoints class. Until those classes are ready, we just use
        # pymatgen here.
        sym_prec = self.incar.get("SYMPREC", 1e-5) if self.incar else 1e-5
        kpath = HighSymmKpath(structure_cleaned, symprec=sym_prec)
        frac_k_points, k_points_labels = kpath.get_kpoints(
            line_density=self.kpoints_line_density,
            coords_are_cartesian=False,
        )
        kpoints = Kpoints(
            comment="Non SCF run along symmetry lines",
            style=Kpoints.supported_modes.Reciprocal,
            num_kpts=len(frac_k_points),
            kpts=frac_k_points,
            labels=k_points_labels,
            kpts_weights=[1] * len(frac_k_points),
        )
        kpoints.write_file(os.path.join(directory, "KPOINTS"))
        ##############

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure_cleaned.composition.elements,
            self.functional,
            os.path.join(directory, "POTCAR"),
            self.potcar_mappings,
        )

    def _write_output_summary(self, directory, vasprun):
        """
        In addition to writing the normal VASP output summary, this also plots
        the bandstructure to "band_structure.png"
        """

        # run the normal output
        super()._write_output_summary(directory, vasprun)

        bs_plotter = BSPlotter(vasprun.get_band_structure(line_mode=True))
        plot = bs_plotter.get_plot()
        plot_filename = os.path.join(directory, "band_structure.png")
        plot.savefig(plot_filename)



# -*- coding: utf-8 -*-

from simmate.workflow_engine.workflow import (
    Workflow,
    Parameter,
    ModuleStorage,
)

from simmate.calculators.vasp.workflows.relaxation.quality_00 import (
    workflow as relaxation_quality00,
)
from simmate.calculators.vasp.workflows.relaxation.quality_01 import (
    workflow as relaxation_quality01,
)
from simmate.calculators.vasp.workflows.relaxation.quality_02 import (
    workflow as relaxation_quality02,
)
from simmate.calculators.vasp.workflows.relaxation.quality_03 import (
    workflow as relaxation_quality03,
)
from simmate.calculators.vasp.workflows.relaxation.quality_04 import (
    workflow as relaxation_quality04,
)
from simmate.calculators.vasp.workflows.energy.quality_04 import (
    workflow as energy_quality04,
)

from simmate.calculators.vasp.database.relaxation import StagedRelaxation
from simmate.calculators.vasp.database.energy import Quality04StaticEnergy


# init workflow tasks
# OPTIMIZE: Make this a for-loop in Prefect 2.0! We can use a for-loop in the
# workflow context below too.
relax_task_00 = relaxation_quality00.to_workflow_task()
relax_task_01 = relaxation_quality01.to_workflow_task()
relax_task_02 = relaxation_quality02.to_workflow_task()
relax_task_03 = relaxation_quality03.to_workflow_task()
relax_task_04 = relaxation_quality04.to_workflow_task()
static_task_04 = energy_quality04.to_workflow_task()

with Workflow("relaxation/staged") as workflow:

    structure = Parameter("structure")
    command = Parameter("command", default="vasp_std > vasp.out")

    # Our first relaxation is directly from our inputs. The remaining one
    # pass along results
    run_id_00 = relax_task_00(
        structure=structure,
        command=command,
    )

    # TODO: Use a for-loop in Prefect 2.0!

    # relaxation 01
    run_id_01 = relax_task_01(
        structure={
            "calculation_table": "Quality00Relaxation",
            "directory": run_id_00["directory"],
            "structure_field": "structure_final",
        },
        command=command,
    )

    # relaxation 02
    run_id_02 = relax_task_02(
        structure={
            "calculation_table": "Quality01Relaxation",
            "directory": run_id_01["directory"],
            "structure_field": "structure_final",
        },
        command=command,
    )

    # relaxation 03
    run_id_03 = relax_task_03(
        structure={
            "calculation_table": "Quality02Relaxation",
            "directory": run_id_02["directory"],
            "structure_field": "structure_final",
        },
        command=command,
    )

    # relaxation 04
    run_id_04 = relax_task_04(
        structure={
            "calculation_table": "Quality03Relaxation",
            "directory": run_id_03["directory"],
            "structure_field": "structure_final",
        },
        command=command,
    )

    # Static Energy (same quality as 04 relaxation)
    run_id_05 = static_task_04(
        structure={
            "calculation_table": "Quality04Relaxation",
            "directory": run_id_04["directory"],
            "structure_field": "structure_final",
        },
        command=command,
    )

workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-Relaxation"
workflow.calculation_table = StagedRelaxation
workflow.result_table = Quality04StaticEnergy
workflow.register_kwargs = ["prefect_flow_run_id"]
workflow.result_task = run_id_05
workflow.s3tasks = [
    w.s3task
    for w in [
        relaxation_quality00,
        relaxation_quality01,
        relaxation_quality02,
        relaxation_quality03,
        relaxation_quality04,
        energy_quality04,
    ]
]

workflow.__doc__ = """
    Runs a series of increasing-quality relaxations and then finishes with a single
    static energy calculation.
    
    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/quality00
        - relaxation/quality01
        - relaxation/quality02
        - relaxation/quality03
        - relaxation/quality04
        - static-energy/quality04
    
    This workflow is most useful for randomly-created structures or extremely 
    large supercells. More precise relaxations+energy calcs should be done 
    afterwards because ettings are still below MIT and Materials Project quality.
"""



# -*- coding: utf-8 -*-

import prefect
from prefect import Task

from simmate.utilities import get_directory


class LoadNestedCalculationTask(Task):
    def __init__(self, calculation_table, **kwargs):
        self.calculation_table = calculation_table
        super().__init__(**kwargs)

    def run(self, directory):

        # first grab the calculation entry
        calc = self.calculation_table.from_prefect_id(prefect.context.flow_run_id)

        # even though S3Task creates a directory when passed None, it is useful
        # to make it here first because some workflows require the folder name between
        # each calculation (see workflows.relaxation.staged for an example). We therefore
        # make the directory upfront!
        new_directory = get_directory(directory)

        # set the directory for the calculation
        calc.directory = get_directory(new_directory)
        calc.save()

        return new_directory



# -*- coding: utf-8 -*-

from simmate.configuration.django import setup_full  # sets database connection
from simmate.database.prototypes.aflow import AflowPrototype

from django_pandas.io import read_frame


class AflowPrototypeStructures:

    """
    Given a compositon, this class can be used to create all structures from
    the prototypes in the AFLOW library.

    At the moment, this does not allow for disordered versions of prototypes.
    This would be like using a supercell of a 1-site prototype for a 2-site target
    composition. A concrete example of this is using the NaCl rocksalt prototype
    for a NaClBr or NaKClBr compositions -- each of these compositions would
    require a supercell of the NaCl prototype structure becuase they need more
    than 2 sites.
    """

    def __init__(self, composition, allow_multiples=True, max_sites=115):
        # composition --> pymatgen composition object (ideally a reduced composition)
        # allow_multiples --> whether to allow structures where the reduced
        # compositions are the same, but the actual number of sites differs.
        # For example, Ca6N3 would be allow to return structures that give
        # Ca2N, Ca8N4, and so forth.
        # max_sites defaults to 115 because this would include all prototypes
        # in the aflow database (current largest is (105 sites))

        # First let's load all of the prototype structures that have a compatible
        # number of sites in the structure.

        # There are two conditions that we need to meet for prototypes here:
        #   (1) the number of wyckoff sites is compatible with nelements OR nsites
        #   (2) the number of individual sites is compatible with nsites
        # This is because we can only insert at symmetrically unique sites (bc
        # we ignore disoreder superlattices here) and the total number of sites
        # after symmetry operations needs to match our target composition.
        # Note that nsites is *not* always a multiple of nsites_wyckoff! For
        # example the Spinel prototype has nsites=14 and nsites_wyckoff=3.

        # If we allow for multiples then  want all multiples nsites in our
        # desired composition. For example if our compositon, has 6 sites, we
        # would pull all prototypes that have 6, 12, 19, ..., 6*N sites.
        if allow_multiples:
            # First generate a list of the allow number of sites
            # We use 40 here because this the maximum number of unique wyckoff
            # sites of any prototype (actual max is 36).
            target_nsites_wyckoff_list = [
                int(composition.num_atoms * n)
                for n in range(1, (40 // len(composition.elements)) + 1)
            ]
            target_nsites_list = [
                int(composition.num_atoms * n)
                for n in range(1, (max_sites // int(composition.num_atoms)) + 1)
            ]
            # Now filter off the prototypes that have the proper number of sites
            prototype_entries = AflowPrototype.objects.filter(
                nsites_wyckoff__in=target_nsites_list,
                nsites__in=target_nsites_list,
            ).all()

        # If we don't want multiples, we can just filter by
        # an exact match for number of total sites and wyckoff sites
        else:
            prototype_entries = AflowPrototype.objects.filter(
                nsites_wyckoff=int(composition.elements),
                nsites=int(composition.num_atoms),
            ).all()

        # for user convience, let's store these selected prototypes as a dataframe
        # and attach it to this class instance
        self.prototype_dataframe = read_frame(prototype_entries)

        # Now that we have the prototype entries, let's convert them to
        # structure objects for us to use below.
        self.prototype_structures = [
            prototype.to_toolkit() for prototype in prototype_entries
        ]



##############################################################################

# List of MatMiner featurizers and (TO-DO) whether they're good for fingerprinting
# from matminer.featurizers.structure import
#
# DensityFeatures
# Dimensionality
# ElectronicRadialDistributionFunction
# GlobalSymmetryFeatures
# RadialDistributionFunction
# PartialRadialDistributionFunction
# CoulombMatrix
# SineCoulombMatrix
# OrbitalFieldMatrix
# MinimumRelativeDistances
# EwaldEnergy
# BondFractions
# BagofBonds
# StructuralHeterogeneity
# MaximumPackingEfficiency
# ChemicalOrdering
# StructureComposition
# XRDPowderPattern
# CGCNNFeaturizer
# JarvisCFID
# SOAP
# GlobalInstabilityIndex
# StructuralComplexity
# SiteStatsFingerprint
#       from matminer.featurizers.site import
#       CrystalNNFingerprint
#       AGNIFingerprints
#       OPSiteFingerprint
#       VoronoiFingerprint
#       IntersticeDistribution
#       ChemicalSRO
#       GaussianSymmFunc
#       EwaldSiteEnergy
#       ChemEnvSiteFingerprint
#       CoordinationNumber
#       GeneralizedRadialDistributionFunction
#       AngularFourierSeries
#       LocalPropertyDifference
#       BondOrientationalParameter
#       SiteElementalProperty
#       AverageBondLength
#       AverageBondAngle

##############################################################################

#!!! THIS SHOULD BE MOVED TO MATMINER!! Speak with their devs on github

import numpy as np

from matminer.featurizers.base import BaseFeaturizer
from matminer.featurizers.site import (
    OPSiteFingerprint,
    CoordinationNumber,
    LocalPropertyDifference,
    AverageBondAngle,
    AverageBondLength,
    CrystalNNFingerprint,
)
from matminer.featurizers.utils.stats import PropertyStats
from pymatgen.analysis.local_env import VoronoiNN


class PartialsSiteStatsFingerprint(BaseFeaturizer):
    """
    Computes statistics of properties across all sites in a structure, and
    breaks these down by element. This featurizer first uses a site featurizer
    class (see site.py for options) to compute features of each site of a
    specific element in a structure, and then computes features of the entire
    structure by measuring statistics of each attribute.
    Features:
        - Returns each statistic of each site feature, broken down by element
    """

    def __init__(
        self,
        site_featurizer,
        stats=("mean", "std_dev"),
        elements=None,
        covariance=False,
    ):
        """
        Args:
            site_featurizer (BaseFeaturizer): a site-based featurizer
            stats ([str]): list of weighted statistics to compute for each feature.
                If stats is None, a list is returned for each features
                that contains the calculated feature for each site in the
                structure.
                *Note for nth mode, stat must be 'n*_mode'; e.g. stat='2nd_mode'
            elements ([str]): list of elements to include. Default is all.
            covariance (bool): Whether to compute the covariance of site features
        """

        self.site_featurizer = site_featurizer
        self.stats = tuple([stats]) if type(stats) == str else stats
        if self.stats and "_mode" in "".join(self.stats):
            nmodes = 0
            for stat in self.stats:
                if "_mode" in stat and int(stat[0]) > nmodes:
                    nmodes = int(stat[0])
            self.nmodes = nmodes

        self.elements_ = elements
        self.covariance = covariance

    def fit(self, X, y=None):
        """Define the list of elements to be included in the PRDF. By default,
        the PRDF will include all of the elements in `X`
        Args:
            X: (numpy array nx1) structures used in the training set. Each entry
                must be Pymatgen Structure objects.
            y: *Not used*
            fit_kwargs: *not used*
        Returns:
            self
        """

        elements = []
        for s in X:
            s_elements = [e.symbol for e in s.composition.elements]
            for e in s_elements:
                if e not in elements:
                    elements.append(e)

        # Store the elements
        self.elements_ = elements

        return self

    @property
    def _site_labels(self):
        return self.site_featurizer.feature_labels()

    def featurize(self, s):
        """
        Get PSSF of the input structure.
        Args:
            s: Pymatgen Structure object.
        Returns:
            pssf: 1D array of each element's ssf
        """

        if self.elements_ is None:
            raise Exception("You must run 'fit' first!")

        output = []
        for e in self.elements_:
            pssf_stats = self.compute_pssf(
                s, e
            )  # Assemble the PSSF for the given element
            output.append(pssf_stats)

        return np.hstack(output)

    def compute_pssf(
        self, s, e
    ):  #!! THIS IS FOR A SET ANALYSIS - CHANGE TO SINGLE ELEMENT
        # Get each feature for each site
        vals = [[] for t in self._site_labels]
        for i, site in enumerate(s.sites):
            if site.specie.symbol == e:
                opvalstmp = self.site_featurizer.featurize(s, i)
                for j, opval in enumerate(opvalstmp):
                    if opval is None:
                        vals[j].append(0.0)
                    else:
                        vals[j].append(opval)

        # If the user does not request statistics, return the site features now
        if self.stats is None:
            return vals

        # Compute the requested statistics
        stats = []
        for op in vals:
            for stat in self.stats:
                stats.append(PropertyStats().calc_stat(op, stat))

        # If desired, compute covariances
        if self.covariance:
            if len(s) == 1:
                stats.extend([0] * int(len(vals) * (len(vals) - 1) / 2))
            else:
                covar = np.cov(vals)
                tri_ind = np.triu_indices(len(vals), 1)
                stats.extend(covar[tri_ind].tolist())

        return stats

    def feature_labels(self):  #!!!! CHANGE THIS TO ACCOUNT FOR MULTIPLE ELEMENTS
        if self.stats:
            labels = []
            # Make labels associated with the statistics
            for attr in self._site_labels:
                for stat in self.stats:
                    labels.append("%s %s" % (stat, attr))

            # Make labels associated with the site labels
            if self.covariance:
                sl = self._site_labels
                for i, sa in enumerate(sl):
                    for sb in sl[(i + 1) :]:
                        labels.append("covariance %s-%s" % (sa, sb))
            return labels
        else:
            return self._site_labels

    def citations(self):
        return self.site_featurizer.citations()

    def implementors(self):
        return ["Jack D. Sundberg"]

    @staticmethod
    def from_preset(preset, **kwargs):
        """
        Create a PartialsSiteStatsFingerprint class according to a preset
        Args:
            preset (str) - Name of preset
            kwargs - Options for PartialsSiteStatsFingerprint
        """

        if preset == "CrystalNNFingerprint_cn":
            return PartialsSiteStatsFingerprint(
                CrystalNNFingerprint.from_preset("cn", cation_anion=False), **kwargs
            )

        elif preset == "CrystalNNFingerprint_cn_cation_anion":
            return PartialsSiteStatsFingerprint(
                CrystalNNFingerprint.from_preset("cn", cation_anion=True), **kwargs
            )

        elif preset == "CrystalNNFingerprint_ops":
            return PartialsSiteStatsFingerprint(
                CrystalNNFingerprint.from_preset("ops", cation_anion=False), **kwargs
            )

        elif preset == "CrystalNNFingerprint_ops_cation_anion":
            return PartialsSiteStatsFingerprint(
                CrystalNNFingerprint.from_preset("ops", cation_anion=True), **kwargs
            )

        elif preset == "OPSiteFingerprint":
            return PartialsSiteStatsFingerprint(OPSiteFingerprint(), **kwargs)

        elif preset == "LocalPropertyDifference_ward-prb-2017":
            return PartialsSiteStatsFingerprint(
                LocalPropertyDifference.from_preset("ward-prb-2017"),
                stats=["minimum", "maximum", "range", "mean", "avg_dev"],
            )

        elif preset == "CoordinationNumber_ward-prb-2017":
            return PartialsSiteStatsFingerprint(
                CoordinationNumber(
                    nn=VoronoiNN(weight="area"), use_weights="effective"
                ),
                stats=["minimum", "maximum", "range", "mean", "avg_dev"],
            )

        elif preset == "Composition-dejong2016_AD":
            return PartialsSiteStatsFingerprint(
                LocalPropertyDifference(
                    properties=[
                        "Number",
                        "AtomicWeight",
                        "Column",
                        "Row",
                        "CovalentRadius",
                        "Electronegativity",
                    ],
                    signed=False,
                ),
                stats=["holder_mean::%d" % d for d in range(0, 4 + 1)] + ["std_dev"],
            )

        elif preset == "Composition-dejong2016_SD":
            return PartialsSiteStatsFingerprint(
                LocalPropertyDifference(
                    properties=[
                        "Number",
                        "AtomicWeight",
                        "Column",
                        "Row",
                        "CovalentRadius",
                        "Electronegativity",
                    ],
                    signed=True,
                ),
                stats=["holder_mean::%d" % d for d in [1, 2, 4]] + ["std_dev"],
            )

        elif preset == "BondLength-dejong2016":
            return PartialsSiteStatsFingerprint(
                AverageBondLength(VoronoiNN()),
                stats=["holder_mean::%d" % d for d in range(-4, 4 + 1)]
                + ["std_dev", "geom_std_dev"],
            )

        elif preset == "BondAngle-dejong2016":
            return PartialsSiteStatsFingerprint(
                AverageBondAngle(VoronoiNN()),
                stats=["holder_mean::%d" % d for d in range(-4, 4 + 1)]
                + ["std_dev", "geom_std_dev"],
            )

        else:
            # TODO: Why assume coordination number? Should this just raise an error? - lw
            # One of the various Coordination Number presets:
            # MinimumVIRENN, MinimumDistanceNN, JmolNN, VoronoiNN, etc.
            try:
                return PartialsSiteStatsFingerprint(
                    CoordinationNumber.from_preset(preset), **kwargs
                )
            except:
                pass

        raise ValueError("Unrecognized preset!")



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

import os

# First we need to load the default configuration for Dask. This is located
# in the simmate configuration directory under the name dask_cluster.yaml.
# This needs to be done BEFORE we import any dask modules, which is why this is
# in the init.
os.environ.setdefault(
    "DASK_CONFIG",
    os.path.join(os.path.expanduser("~"), "simmate", "dask_cluster.yaml"),
)
# TODO: I need to update this to allow for other directories

from .client import get_dask_client
from .submit import batch_submit



# -*- coding: utf-8 -*-

import os
import platform
import time
import signal
import subprocess
import yaml

import pandas

import prefect
from prefect.core.task import Task
from prefect.utilities.tasks import defaults_from_attrs

from simmate.utilities import get_directory, make_archive

from typing import List, Any
from simmate.toolkit import Structure
from simmate.workflow_engine.error_handler import ErrorHandler

# cleanup_on_fail=False, # TODO I should add a Prefect state_handler that can
# reset the working directory between task retries -- in some cases we may
# want to delete the entire directory.

# OPTIMIZE: I think this class would greatly benefit from asyncio so that we
# know exactly when a shelltask completes, rather than looping and checking every
# set timestep.
# https://docs.python.org/3/library/asyncio-subprocess.html#asyncio.create_subprocess_exec


class S3Task(Task):
    """
    The Supervised-Staged-Shell Task (aka "S3Task")
    -----------------------------------------------

    This class contains the core functionality to **supervise** a **staged** task
    involving some **shell** command.

    Let's breakdown what this means...

    A *shell* command is a single call to some external program. For example,
    VASP requires that we call the "vasp_std > vasp.out" command in order to run a
    calculation. We consider calling external programs a *staged* task made
    up of three steps:

    - setup = writing any input files required for the program
    - execution = actually calling the command and running our program
    - workup = loading data from output files back into python

    And for *supervising* the task, this means we monitor the program while the
    execution stage is running. So once a program is started, Simmate can check
    output files for common errors/issues. If an error is found, we stop the
    program, fix the issue, and then restart it. Any fixes that were made are
    written to "simmate_corrections.csv".


    <!--
    TODO: Make a simple diagram to visualize the overall process and add it here.
    It will be similar to Custodian's, but we don't have a list of jobs here.
    https://materialsproject.github.io/custodian/index.html#usage
    The steps are...

    - Write Input Files based on custom+defualt settings
    - Run the calculation by calling the program
    - Load ouput files
    - check for errors
    - [correct them, rerun]
    - postprocess/analysis
    -->

    This entire process (the different stages and monitoring) is carried out
    using the ``run()`` method. You rarely use this class directly. Instead,
    you typically use a subclass of it. As a user, you really just need to do
    something like this:

    ``` python
       from simmate.calculator.example.tasks import ExampleTask

       my_task = ExampleTask()
       my_result = my_task.run()
    ```

    And that's it!

    For experts, this class can be viewed as a combination of prefect's ShellTask,
    a custodian Job, and Custodian monitoring. When subclassing this, we can absorb
    functionality of pymatgen.io.vasp.sets too. By merging all of these together
    into one class, we make things much easier for users and creating new Tasks.


    Inheriting from this class
    --------------------------

    This class is commonly used to make tasks for our calculator modules, so you
    will likely want to subclass this. Here is a basic example of inheriting
    and then running a task:

    ``` python

    from simmate.workflow_engine import S3Task
    from example.error_handlers import PossibleError1, PossibleError2


    class ExampleTask(SSSTask):

        command = "echo example"  # just prints out "example"
        max_corrections = 7
        error_handlers = [PossibleError1, PossibleError2]
        polling_timestep = 0.1
        monitor_freq = 10
        some_new_setting = 123

        def setup(self, structure, directory):  # <-- MUST have these two args
            print("I'm setting things up!")
            print(f"My new setting is {some_new_setting}")

        def workup(self, directory):  # <-- MUST have this arg
            print("I'm working things up!")


    task = ExampleTask()
    result = task.run()
    ```

    There are a couple things to note here:

    - It's optional to set/overwrite attributes. You can also add new ones too.
    - It's optional to write a new __init__, setup, or workup methods
    - make sure you include the structure/directory inputs, even if you don't use them.
    - Don't add new kwargs to methods. Instead handle these options through attributes.

    For a full (and advanced) example, of a subclass take a look at
    `simmate.calculators.vasp.tasks.base.VaspTask`.

    """

    # I set this here so that I don't have to copy/paste the init method
    # every time I inherit from this class and want to update the default
    # command to use for the child class.
    command: str = None
    """
    The defualt shell command to use.
    """

    # While it's not needed for a number of cases, it's extremely common for the
    # setup method to need an input structure in matsci. I therefore include
    # this rather than having a nearly identical subclass that could cause
    # some confusion.
    requires_structure: bool = False
    """
    Indicates whether a structure is needed if for the run() method.
    """

    error_handlers: List[ErrorHandler] = []
    """
    A list of ErrorHandler objects to use in order of priority (that is, highest
    priority is first). If one handler is triggered, the correction will be 
    applied and none of the following handlers will be checked.
    """

    def __init__(
        self,
        max_corrections: int = 5,
        monitor: bool = True,
        polling_timestep: float = 1,
        monitor_freq: int = 300,
        save_corrections_to_file: bool = True,
        corrections_filename: str = "simmate_corrections.csv",
        compress_output: bool = False,
        **kwargs: Any,
    ):
        """
        Creates a task instance of this class. The parameters passed will be the
        same every time you call the task.run() method.

        Note, many of the parameters here are for the Monitoring settings. These
        are only ever relevent if there are ErrorHandlers added that have
        is_monitor=True. These handlers run while the shelltask itself is also
        running. Read more about ErrorHandlers for more info.

        #### Parameters

        - `max_corrections`:
            The maximum number of times we can apply a correction and retry the shell
            command. The maximum number of times that corrections will be made (and
            shell command reran) before giving up on the calculation. Note, once this
            limit is exceeded, the error is stored without correcting or restarting
            the run.
        - `monitor`:
            Whether to run monitor handlers while the command runs. False means
            wait until the job has completed.
        - `polling_timestep`:
            If we are monitoring the job for errors while it runs, this is how often
            (in seconds) we should check the status of our job. Note this check is
            just whether the job is done or not. This is NOT how often we check for
            errors. See monitor_freq for that.
        - `monitor_freq`:
            The frequency we should run check for errors with our monitors. This is
            based on the polling_timestep loops. For example, if we have a
            polling_timestep of 10 seconds and a monitor_freq of 2, then we would run
            the monitor checks every other loop -- or every 2x10 = 20 seconds. The
            default values of polling_timestep=10 and monitor_freq=30 indicate that
            we run monitoring functions every 5 minutes (10x30=300s=5min).
        - `save_corrections_to_file`:
            Whether to write a log file of the corrections made. The default is True.
        - `corrections_filename`:
            If save_corrections_to_file is True, this is the filename of where
            to write the corrections. The default is "simmate_corrections.csv".
        - `compress_output`:
            Whether to compress the directory to a zip file at the end of the
            task run. After compression, it will also delete the directory.
            The default is False.
        - `**kwargs`:
            All extra arguments supported by
            [prefect.core.task.Task](https://docs.prefect.io/api/latest/core/task.html).
        """

        # Save the provided settings
        self.monitor = monitor
        self.max_corrections = max_corrections
        self.polling_timestep = polling_timestep
        self.monitor_freq = monitor_freq
        self.compress_output = compress_output
        self.save_corrections_to_file = save_corrections_to_file
        self.corrections_filename = corrections_filename

        # now inherit the parent Prefect Task class
        super().__init__(**kwargs)

    def setup(self, structure: Structure, directory: str, **kwargs):
        """
        This abstract method is ran before the command is actually executed. This
        allows for some pre-processing, such as writing input files or any other
        analysis.

        You should never call this method directly unless you are debugging. This
        is becuase setup() is normally called within the run() method.

        Some tasks don't require a setup() method, so by default, this method
        doesn't nothing but "pass".

        #### Parameters

        - `structure`:
            The structure to use for the task, if one is required.
        - `directory`:
            The directory to run everything in. Must exist already.
        - `**kwargs`:
            Extra kwargs that may be passed to some function within. Because
            Simmate prefers fixed settings for their workflows, this is typically
            not used, but instead, keywords should be explicitly defined when
            writing a setup method.
        """
        # Be sure to include directory and structure (or **kwargs) as input
        # arguments for higher-level compatibility with the run method.
        pass

    def execute(self, directory: str, command: str):
        """
        This calls the command within the target directory and handles all error
        handling as well as monitoring of the job.

        You should never call this method directly unless you are debugging. This
        is becuase `execute` is normally called within the `run` method.

        Some tasks don't require a setup() method, so by default, this method
        does nothing but "pass".

        #### Parameters

        - `directory`:
            The directory to run everything in.
        - `command`:
            The command that will be called during execution.

        #### Returns

        - `corrections`
            A list of tuples where each entry is a error identified and the
            correction applied. Ex: [("ExampleError", "ExampleCorrection")]

        """

        # some error_handlers run while the shelltask is running. These are known as
        # Monitors and are labled via the is_monitor attribute. It's good for us
        # to separate these out from other error_handlers.
        self.monitors = [
            handler for handler in self.error_handlers if handler.is_monitor
        ]

        # We start with zero corrections that we slowly add to. This can be
        # thought of as a table with headers of...
        #   ("applied_errorhandler", "correction_applied")
        # NOTE: I took out the table headers and may add them back in
        corrections = []

        # ------ start of main while loop ------

        # we can try running the shelltask up to max_corrections. Because only one
        # correction is applied per attempt, you can view this as the maximum
        # number of attempts made on the calculation.
        while len(corrections) <= self.max_corrections:

            # launch the shelltask without waiting for it to complete. Also,
            # make sure to use common shell commands and to set the working
            # directory.
            #
            # Stderr keyword indicates that we should capture the error if one
            # occurs so that we can report it to the user.
            #
            # The preexec_fn keyword allows us to properly terminate jobs that
            # are launched with parallel processes (such as mpirun). This assigns
            # a parent id to it that we use when killing a job (if an error
            # handler calls for us to do so). This isn't possible on Windows though.
            #
            # OPTIMIZE / BUG: preexec_fn adds about 0.02s overhead to the calculation
            # so we may not want to always use it... Instead we could try to only
            # use it when the command includes "mpirun". Though this may introduce
            # a bug if another is another parallel command used besides mpirun.
            # An example of this might be deepmd which automatically submits
            # things in parallel without calling mpirun up-front.
            process = subprocess.Popen(
                command,
                cwd=directory,
                shell=True,
                preexec_fn=None if platform.system() == "Windows"
                # or "mpirun" not in command  # See bug/optimize comment above
                else os.setsid,
                stderr=subprocess.PIPE,
            )

            # Assume the shelltask has no errors until proven otherwise
            has_error = False

            # If monitor=True, then we want to supervise this shelltask as it
            # runs. If montors=[m1,m2,...], then we have monitors in place to actually
            # perform the monitoring. If both of these cases are true, then we
            # want to go through the error_handlers to check for errors until
            # the shelltask completes.
            if self.monitor and self.monitors:

                # ------ start of monitor while loop ------

                # We want to loop until we find an error and keep track of
                # which loops to run the monitor function on because we don't
                # want them to run nonstop. This variable allows us to monitor
                # checks every Nth loop, while we check the shelltask status
                # on all other loops.
                monitor_freq_n = 0
                while not has_error:
                    monitor_freq_n += 1
                    # Sleep the set amount before checking the shelltask status
                    time.sleep(self.polling_timestep)

                    # check if the shelltasks is complete. poll will return 0
                    # when it's done, in which case we break the loop
                    if process.poll() is not None:
                        break
                    # check whether we should run monitors on this poll loop
                    if monitor_freq_n % self.monitor_freq == 0:
                        # iterate through each monitor
                        for error_handler in self.monitors:
                            # check if there's an error with this error_handler
                            # and grab the error if so
                            error = error_handler.check(directory)
                            if error:
                                # determine if it is_terminating
                                if error_handler.is_terminating:
                                    # If so, we kill the process but don't apply
                                    # the fix quite yet. That step is done below.
                                    self._terminate_job(process, command)
                                # Otherwise apply the fix and let the shelltask end
                                # naturally. An example of this is for codes
                                # where you add a STOP file to get it to
                                # finish rather than just killing the process.
                                # This is the special case error_handler that I talk
                                # about in my notes, where we really want to
                                # end the shelltask right away.
                                else:
                                    # apply the fix now
                                    correction = error_handler.correct(directory)
                                    # record what's been changed
                                    corrections.append((error_handler.name, correction))
                                # there's no need to look at the other monitors
                                # so break from the for-loop. We also don't
                                # need to monitor the stagedtask anymore since we just
                                # terminated it or signaled for its graceful
                                # end. So update the while-loop condition.
                                has_error = True
                                break
                # ------ end of monitor while loop ------
            # Now just wait for the process to finish. Note we use communicate
            # instead of the .wait() method. This is the recommended method
            # when we have stderr=subprocess.PIPE, which we use above.
            output, errors = process.communicate()
            # check if the return code is non-zero and thus failed.
            # The 'not has_error' is because terminate() will give a nonzero
            # when a monitor is triggered. We don't want to raise that
            # exception here but instead let the monitor handle that
            # error in the code below.
            if process.returncode != 0 and not has_error:
                # convert the error from bytes to a string
                errors = errors.decode("utf-8")
                # and report the error to the user
                raise NonZeroExitError(
                    f"The command ({command}) failed. The error output was...\n {errors}"
                )
            # Check for errors again, because a non-monitor may be higher
            # priority than the monitor triggered above (if there was one).
            # Since the error_handlers are in order of priority, only the first
            # will actually be applied and then we can retry the calc.
            for error_handler in self.error_handlers:

                # NOTE - The following special case is handled above:
                #   error_handler.is_monitor and not error_handler.is_terminating
                # BUG: I can see this being a source of bugs in the process so I
                # need to reconsider subclassing this special case. For now,
                # users should have this case at the lowest priority.

                # check if there's an error with this error_handler and grab the
                # error if there is one
                error = error_handler.check(directory)
                if error:
                    # record the error in case it wasn't done so above
                    has_error = True
                    # And apply the proper correction if there is one.
                    # Some error_handlers will even raise an error here signaling
                    # that the stagedtask is unrecoverable and a lost cause.
                    correction = error_handler.correct(directory)
                    # record what's been changed
                    corrections.append((error_handler.name, correction))
                    # break from the error_handler for-loop as we only apply the
                    # highest priority fix and nothing else.
                    break
            # write the log of corrections to file if requested. This is written
            # as a CSV file format and done every while-loop cycle because it
            # lets the user monitor the calculation and error handlers applied
            # as it goes. If no corrections were applied, we skip writing the file.
            if self.save_corrections_to_file and corrections:
                # compile the corrections metadata into a dataframe
                data = pandas.DataFrame(
                    corrections,
                    columns=["error_handler", "correction_applied"],
                )
                # write the dataframe to a csv file
                data.to_csv(
                    os.path.join(directory, self.corrections_filename),
                    index=False,
                )
            # now that we've gone through the error_handlers, let's see if any
            # non-terminating errors were found (terminating ones would raise
            # an error above). If there are no errors, we've finished the
            # calculation and can exit the while loop. Otherwise, just leave
            # everything where it's at and restart the while-loop with a new
            # attempt
            if not has_error:
                # break the while-loop
                break
        # ------ end of main while loop ------

        # make sure the while loop didn't exit because of the correction limit
        if len(corrections) >= self.max_corrections:
            raise MaxCorrectionsError(
                "The number of maximum corrections has been exceeded. Note the final "
                "error and its fix are still listed in the corrections file, but it "
                "was never used."
            )
        # now return the corrections for them to stored/used elsewhere
        return corrections

    @staticmethod
    def _terminate_job(process: subprocess.Popen, command: str):
        """
        Stopping the command we submitted can be a tricky business if we are running
        scripts in parallel (such as using mpirun). Different computers and OSs
        can require a different 'kill' command so we establish this separate
        method that can be overwritten.

        If you know your command needs a special case to kill all of its spawn
        processes, you can overwrite this method as well.

        Users should never call this directly becuase this is instead applied
        within the execute() method.

        #### Parameters

        - `process`:
            The process object that will be terminated.
        - `command`:
            the command used to launch the process. This is sometimes useful
            when searching for all running processes under this name.
        """

        # The operating system (Windows, OSX, or Linux) will give us the best guess
        # as to where to access the blender command. The results are as you'd expect
        # except for Macs, which return "Darwin".
        #   Linux: Linux
        #   Mac: Darwin
        #   Windows: Windows
        operating_system = platform.system()

        # The normal line to end a popen process is just...
        #   process.terminate()
        # However this struggles to kill all "child" processes if we are using
        # something like mpirun to run things in parallel. Instead, we use the
        # os module to grab the parent id and send that the termination signal,
        # which is also passed on to all child processes.
        # This command also doesn not work on windows, so I need to address this
        # as well.
        if operating_system != "Windows":
            os.killpg(os.getpgid(process.pid), signal.SIGKILL)
        # note: SIGTERM is the normal signal but I use SIGKILL to try to address
        # permission errors.
        else:
            process.terminate()

        # As an example of an alternative approach to killing a job, here is
        # what Custodian (Materials Project) tries when killing a VASP job
        # submitted through mpirun:
        # try:
        #     os.system("killall vasp")
        # except Exception:
        #     pass

        # TODO: make it so terminate scripts can be loaded from the user's config
        # directory, which will make it so they don't have to overwrite all the
        # classes that inherit from this one.

    def workup(self, directory: str):
        """
        This method is called at the end of a job, *after* error detection.
        This allows post-processing, such as cleanup, analysis of results,
        etc. This should return the result of the entire job, such as a
        the final structure or final energy calculated.

        You should never call this method directly unless you are debugging. This
        is becuase workup() is normally called within the run() method.

        Some tasks don't require a workup() method, so by default, this method
        does nothing but "pass".

        #### Parameters

        - `directory`:
            The directory to run everything in.

        """
        # Be sure to include directory (or **kwargs) as input argument for
        # higher-level compatibility with the run method and SupervisedStagedTask
        # You should never need to call this method directly!
        pass

    @defaults_from_attrs("command")
    def run(
        self,
        structure: Structure = None,
        directory: str = None,
        command: str = None,
        **kwargs,
    ):
        """
        Runs the entire staged task (setup, execution, workup), which includes
        supervising during execution.

        Call this method once you have your task initialized. For each run you
        can provide a new structure, directory, or command. For example,

        ``` python
        from simmate.calculator.example.tasks import ExampleTask

        my_task = ExampleTask()
        my_result = my_task.run(structure=my_structure, command=my_command)
        ```

        #### Parameters

        - `structure`:
            The structure to use for the task, if one is required.
        - `command`:
            The command that will be called during execution.
        - `directory`:
            The directory to run everything in. This is passed to the ulitities
            function simmate.ulitities.get_directory
        - `**kwargs`:
            Any extra keywords that should be passed to the setup() method.

        Returns
        -------
        - a dictionary of the result, corrections, and working directory used
        for this task run

        """

        # because the command is something that is frequently changed at the
        # workflow level, then we want to make it so the user can set it for
        # each unique task.run() call. Otherwise we grab the default from the
        # class attribute

        # make sure a structure was given if it's required
        if not structure and self.requires_structure:
            raise StructureRequiredError("a structure is required as an input")
        # establish the working directory
        directory = get_directory(directory)

        # run the setup stage of the task
        self.setup(structure, directory, **kwargs)

        # run the shelltask and error supervision stages. This method returns
        # a list of any corrections applied during the run.
        corrections = self.execute(directory, command)

        # run the workup stage of the task. This is where the data/info is pull
        # out from the calculation and is thus our "result".
        result = self.workup(directory)

        # if requested, compresses the directory to a zip file and then removes
        # the directory.
        if self.compress_output:
            make_archive(directory)
        # Return our final information as a dictionary
        return {
            "result": result,
            "corrections": corrections,
            "directory": directory,
            "prefect_flow_run_id": prefect.context.flow_run_id
            if hasattr(prefect.context, "flow_run_id")
            else None,  # when not ran within a prefect flow, there won't be an id
        }

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class.

        By default, this will just grab the class's __dict__ attribute but this
        can be overwritten to show only relevent information.
        """
        return dict(cls.__dict__)

    @classmethod
    def print_config(cls):
        """
        Takes the result of get_config and prints it in a yaml format that is
        easier to read.
        """
        config = cls.get_config()
        print(yaml.dump(config))


# Custom errors that indicate exactly what causes the S3Task to exit.


class MaxCorrectionsError(Exception):
    pass


class NonZeroExitError(Exception):
    pass


class StructureRequiredError(Exception):
    pass



# -*- coding: utf-8 -*-

"""
Runs a NEB on all unique pathways within a structure. 

The folder tree looks like...
```
simmate-task-12345/  # determined by simmate.utilities.get_directory
    ├── bulk_relaxation
    ├── bulk_static_energy
    ├── migration_hop_00
    ├── migration_hop_01
    ...
    └── migration_hop_N  # all migration_hop folders have the same structure
        ├── endpoint_relaxation_start
        ├── endpoint_relaxation_end
        ├── 01
        ├── 02
        ├── 03
        ...
        └── N  # corresponds to image number
```
"""

import os

from prefect import apply_map, unmapped

from simmate.workflow_engine.workflow import (
    Workflow,
    task,
    Parameter,
    ModuleStorage,
)
from simmate.workflow_engine.common_tasks import (
    LoadInputAndRegister,
    parse_multi_command,
)
from simmate.calculators.vasp.workflows.nudged_elastic_band.utilities import (
    BuildDiffusionAnalysisTask,
)
from simmate.calculators.vasp.workflows.relaxation import (
    mit_workflow as relaxation_mit_workflow,
)
from simmate.calculators.vasp.workflows.energy import (
    mit_workflow as energy_mit_workflow,
)
from simmate.calculators.vasp.workflows.nudged_elastic_band.single_path import (
    workflow as neb_workflow,
)


from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
)

workflow_name = "diffusion/all-paths"

# Convert our workflow objects to task objects
relax_bulk = relaxation_mit_workflow.to_workflow_task()
energy_bulk = energy_mit_workflow.to_workflow_task()
run_neb = neb_workflow.to_workflow_task()

# Extra setup tasks
load_input_and_register = LoadInputAndRegister(
    workflow_name=workflow_name,
    input_obj_name="structure",
)  # TODO: make MigrationHop a calc so we can use calculation_table?

build_db = BuildDiffusionAnalysisTask(MITDiffusionAnalysis)

# ------------------------

# TODO: Prefect isn't able to do this for-loop. I may need to ask them how
# to map this kind of input. Hopefully this changes with Prefect Orion. This is
# the code that would be in the workflow context:
#
# for i, hop_id in enumerate(migration_hop_ids):
#     run_neb(
#         migration_hop={
#             "migration_hop_table": "MITMigrationHop",
#             "migration_hop_id": hop_id,
#         },
#         directory=directory_cleaned
#         + os.path.sep
#         + f"migration_hop_{i}",
#         source="DistinctPathFinder",
#         diffusion_analysis_id=None,
#         migration_hop_id=None,
#         command=subcommands["command_supercell"] + ";" + subcommands["command_neb"]
#     )
#
# Instead I need these hacky task objects and code.


@task
def get_dir_name(number: int, directory: str):
    number_fill = str(number).zfill(2)
    return os.path.join(directory, f"migration_hop_{number_fill}")


def map_neb(migration_hop_id: int, directory: str, subcommands: dict):

    # !!! This should pass the mapping index... not id. Not sure how to
    # enumerate this with Prefect.
    hop_directory = get_dir_name(migration_hop_id, directory)

    mapped_neb_task = run_neb(
        migration_hop={
            "migration_hop_table": "MITMigrationHop",
            "migration_hop_id": migration_hop_id,
        },
        directory=hop_directory,
        # source="DistinctPathFinder",
        diffusion_analysis_id=None,
        migration_hop_id=migration_hop_id,
        command=subcommands["command_supercell"] + ";" + subcommands["command_neb"],
    )
    return mapped_neb_task


# ------------------------


with Workflow(workflow_name) as workflow:

    structure = Parameter("structure")
    migrating_specie = Parameter("migrating_specie")
    source = Parameter("source", default=None)
    directory = Parameter("directory", default=None)
    # assume use_previous_directory=False for this flow

    # I separate these out because each calculation is a very different scale.
    # For example, you may want to run the bulk relaxation on 10 cores, the
    # supercell on 50, and the NEB on 200. Even though more cores are available,
    # running smaller calculation on more cores could slow down the calc.
    command = Parameter("command", default="vasp_std > vasp.out")
    # command list expects three subcommands:
    #   command_bulk, command_supercell, and command_neb
    subcommands = parse_multi_command(
        command,
        commands_out=["command_bulk", "command_supercell", "command_neb"],
    )

    # Load our input and make a base directory for all other workflows to run
    # within for us.
    structure_toolkit, directory_cleaned = load_input_and_register(
        input_obj=structure,
        source=source,
        directory=directory,
        command=command,
        migrating_specie=migrating_specie,
    )

    # Our step is to run a relaxation on the bulk structure and it uses our inputs
    # directly. The remaining one tasks pass on results.
    run_id_00 = relax_bulk(
        structure=structure_toolkit,
        command=subcommands["command_bulk"],
        directory=directory_cleaned + os.path.sep + "bulk_relaxation",
    )

    # A static energy calculation on the relaxed structure. This isn't necessarily
    # required for NEB, but it takes very little time.
    run_id_01 = energy_bulk(
        structure={
            "calculation_table": "MITRelaxation",
            "directory": run_id_00["directory"],
            "structure_field": "structure_final",
        },
        command=subcommands["command_bulk"],
        directory=directory_cleaned + os.path.sep + "bulk_static_energy",
    )

    # This step does NOT run any calculation, but instead, identifies all
    # diffusion pathways and builds the necessary database entries.
    migration_hop_ids = build_db(
        structure={
            "calculation_table": "MITStaticEnergy",
            "directory": run_id_01["directory"],
        },
        migrating_specie=migrating_specie,
        directory=directory_cleaned,
        vacancy_mode=True,  # assumed for now
    )

    # Run NEB single_path workflow for all these.
    apply_map(
        map_neb,  # the task that we are mapping
        migration_hop_id=migration_hop_ids,  # this input is mapped
        directory=unmapped(directory_cleaned),  # this input will be constant
        subcommands=unmapped(subcommands),  # this input will be constant
    )


workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-Diffusion"
# workflow.calculation_table = MITDiffusionAnalysis  # not implemented yet
# workflow.register_kwargs = ["prefect_flow_run_id"]
workflow.result_table = MITDiffusionAnalysis
workflow.s3tasks = [
    relaxation_mit_workflow.s3task,
    energy_mit_workflow.s3task,
] + neb_workflow.s3tasks

workflow.__doc__ = """
    Runs a full diffusion analysis on a bulk crystal structure using NEB.
    
    The bulk structure will be geometry optimized and then 
    `simmate.toolkit.diffusion.DistinctPathFinder` is used to find all
    symmetrically unique migration hops in the structure up until the hops 
    become percolating (>0-D). For each unique hop, the workflow 
    diffusion/single_path is submitted.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/mit
        - static-energy/mit
        - a mini task that identifies unique migration hops
        - (for each hop) diffusion/single-path
    
    If you are running this workflow via the command-line, you can run this 
    with...
    
    ``` bash
    simmate workflows run diffusion/all-paths -s example.cif -c "cmd1; cmd2; cmd3"
    ```
    
    Note, the `-c` here is very important! Here we are passing three commands
    separated by semicolons. Each command is passed to a specific workflow call:
        
        - cmd1 --> used for bulk crystal relaxation and static energy
        - cmd2 --> used for endpoint supercell relaxations
        - cmd3 --> used for NEB
    
    Thus, you can scale your resources for each step. Here's a full -c option:
    
    -c "vasp_std > vasp.out; mpirun -n 12 vasp_std > vasp.out; mpirun -n 70 vasp_std > vasp.out"
"""



# -*- coding: utf-8 -*-

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
from prefect.engine.state import TriggerFailed

from typing import List
from simmate.workflow_engine.workflow import Workflow

# BUG: prefect executor causes issues because there is no aysncio. Therefore the
# "waiting" step means a task will be holding onto two Dask workers. Because of this
# Prefect cloud has been disabled as an executor.
#
# import os
# from pathlib import Path
# import yaml
# def get_default_executor_type():
#     # rather than always specifying the executor type whenever this class is used,
#     # it's much more convenient to set the default value here based on a config
#     # file. We check for this in the file "simmate/extra_applications".
#     # For windows, this would be something like...
#     #   C:\Users\exampleuser\simmate\executor.yaml
#     SIMMATE_DIRECTORY = os.path.join(Path.home(), "simmate")
#     EXECUTOR_YAML = os.path.join(SIMMATE_DIRECTORY, "executor.yaml")
#     if os.path.exists(EXECUTOR_YAML):
#         with open(EXECUTOR_YAML) as file:
#             EXECUTOR_TYPE = yaml.full_load(file)["EXECUTOR_TYPE"]
#     else:
#         EXECUTOR_TYPE = "local"
#     return EXECUTOR_TYPE


class WorkflowTask(Task):
    """
    Many times in Material Science, a workflow is made up of other smaller workflows.
    For example, calculations for bandstructure first involve a structure relaxation
    followed by a static energy calculation -- before the bandstructure is even
    calculated.

    For this reason, we need a way to call these workflows as if they were a task.

    The easiest way to make your Workflow into a WorkflowTask is actually NOT
    calling this class. But instead using the to_workflow_task method:

    ``` python
    from simmate.workflows import example_workflow

    my_task = example_workflow.to_workflow_task()
    ```
    """

    # DEV NOTE:
    # Prefect has a module prefect.tasks.prefect.flow_run which is close to this
    # idea -- but this module requires you to be using Prefect Cloud and have all
    # of the flows registered. We need to account for running flows locally too!
    # Therefore, we had to make this custom task here -- where it decided whether
    # we call the workflow.run_cloud or workflow.run method.
    # Consider moving this functionality to Prefect down the line.

    def __init__(
        self,
        workflow: Workflow,
        return_result=True,
        executor_type: str = "local",  # get_default_executor_type(), # BUG
        wait_for_run: bool = True,
        labels: List[str] = [],
        log_stdout: bool = True,
        **kwargs,
    ):
        """
        Creates a Prefect task instance from a workflow and determines how it
        should be ran when task.run() is called.

        Parameters

        - `workflow`:
            The full workflow to run.
        - `return_result`:
            If True, the task run will return the result of the task set as the
            worfklow.result_task attribute. If False, the workflow State is returned.
            The default is True.
        - `executor_type`:
            How the workflow should be scheduled and ran. The options are either
            local or prefect (i.e. prefect cloud). The default is "local".
        - `wait_for_run`:
            If Prefect is used to schedule the workflow, then this indicates
            whether we should wait for the flow to finish or not.
        - `labels`:
            If Prefect is used to schedule the workflow, then this indicates
            what labels should be attached to the flow
        - `log_stdout`:
            Whether to log anything printed by the workflow. The default is True.
        - `**kwargs`:
            All extra arguments supported by
            [prefect.core.task.Task](https://docs.prefect.io/api/latest/core/task.html).
        """

        # If Prefect is used to schedule the workflow, we also may want to
        # To support other Prefect input options. To see all the options, visit...
        # https://docs.prefect.io/api/latest/core/task.html
        # We want to pass the logs from the workflow task up!

        # make sure the user selected a valid option
        if executor_type not in ["local", "prefect"]:
            raise Exception(
                f"{executor_type} is not a valid option! "
                "Please choose between local and prefect"
            )
        # NOTE: We don't support Dask or Simmate executors because this would
        # cause some messy nested parallelization. For example, if the parent
        # workflow is submitted with Dask, this workflow would be submitted as
        # an individual task to that Cluster -- so submitting this child
        # workflow's task to the same Cluster can cause issues. So you should
        # stick to Prefect Cloud for advanced parallelization.

        # Save our inputs
        self.workflow = workflow
        self.return_result = return_result
        self.executor_type = executor_type
        self.wait_for_run = wait_for_run
        self.labels = labels

        # BUG: so far I only support local execution
        if executor_type != "local":
            raise Exception("Only executor_type=local is supported right now.")

        # now inherit the parent Prefect Task class
        # Note we name this task after our attached workflow
        super().__init__(
            name=f"{self.workflow.name}-WorkflowTask",
            log_stdout=log_stdout,
            **kwargs,
        )

    @defaults_from_attrs(
        "executor_type",
        "return_result",
        "wait_for_run",
        "labels",
    )
    def run(
        self,
        executor_type: str = None,
        return_result: bool = None,
        wait_for_run: bool = None,
        labels: List[str] = None,
        **parameters,
    ):
        """
        Runs the workflow as as task.

        #### Parameters

        - `executor_type`:
            How the workflow should be scheduled and ran. The options are either
            local or prefect (i.e. prefect cloud).
        - `return_result`:
            If True, the task run will return the result of the task set as the
            worfklow.result_task attribute. If False, the workflow State is returned.
            The default is True.
        - `wait_for_run`:
            If Prefect is used to schedule the workflow, then this indicates
            whether we should wait for the flow to finish or not.
        - `labels`:
            If Prefect is used to schedule the workflow, then this indicates
            what labels should be attached to the flow
        - `**parameters`:
            All options to pass the workflow.run() method.

        #### Returns

        str
            The flow-run-id if submitted with prefect OR the prefect state if
            the workflow was ran locally

        """
        # The kwargs here are any parameter you would normally pass into the
        # workflow.run() method.

        # If we want this ran locally, we can run the workflow directly. We
        # also make sure the workflow is successful -- otherwise we want to
        # raise an error so that this task is flagged as failed
        if executor_type == "local":
            state = self.workflow.run(**parameters)
            if state.is_failed():
                # Grab the all task errors except for those that never ran
                # because the trigger failed.
                # OPTIMIZE: would the log_stdout=True task option be the better
                # way to forward information? See here:
                #   https://docs.prefect.io/core/concepts/logging.html#logging-stdout
                errors = []
                for task_state in state.result.values():
                    if task_state.is_failed() and not isinstance(
                        task_state, TriggerFailed
                    ):
                        error = task_state.result
                        errors.append(error)

                # TODO: how do you raise a list of errors?
                if len(errors) > 1:
                    raise Exception(
                        "More than one task failed in the workflow and Simmate does "
                        "not currently have a way to raise a list of errors."
                    )
                # otherwise we just have one failure
                else:
                    # we want the error raised so that the task fails
                    raise errors[0]
            # if we reached this point, the workflow was successful and we just return
            # the state as a normal task would

            # Workflows typically return a prefect State object, but when we run
            # workflows as a task, we often want to pass along extra information
            # beyond just some "state". So return_result is True, we return the
            # result of a specific task (determined by flow.result_task).
            if self.return_result and self.workflow.result_task:
                # The prefect api for this is ugly... nothing I can really do
                # about it: https://docs.prefect.io/core/concepts/results.html
                return state.result[self.workflow.result_task].result
            else:
                return state

        # If we are using prefect, we assume that the flow has been registered
        # and simply submit it from here.
        elif executor_type == "prefect":

            raise Exception

            # Now create the flow run in the prefect cloud
            flow_run_id = self.workflow.run_cloud(
                labels,
                wait_for_run,
                **parameters,
            )
            if wait_for_run:
                raise Exception("wait_for_run not implemented for cloud yet")
            return flow_run_id



# -*- coding: utf-8 -*-

import os

from pymatgen.electronic_structure.plotter import DosPlotter

from simmate.calculators.vasp.tasks.energy.materials_project import (
    MatProjStaticEnergy,
)


class MatProjDensityOfStates(MatProjStaticEnergy):
    """
    Calculates the density of states using Materials Project settings.

    This is a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which handles the preliminary relaxation and SCF
    energy calculation for you. This S3Task is only the final step of that workflow.

    See `vasp.workflows.density_of_states`.
    """

    # Settings are based off of pymatgen's NonSCFSet in uniform mode
    incar = MatProjStaticEnergy.incar.copy()
    incar.update(
        ICHARGE=11,
        ISYM=2,
        NEDOS=2001,
    )
    incar.pop("MAGMOM__smart_magmom")

    def _write_output_summary(self, directory, vasprun):
        """
        In addition to writing the normal VASP output summary, this also plots
        the DOS to "density_of_states.png"
        """

        # run the normal output
        super()._write_output_summary(directory, vasprun)

        plotter = DosPlotter()

        # Add the total density of States
        plotter.add_dos("Total DOS", vasprun.complete_dos)

        # add element-projected density of states
        plotter.add_dos_dict(vasprun.complete_dos.get_element_dos())

        # If I want plots for individual orbitals
        # for site in vasprun.final_structure:
        #     spd_dos = vasprun.complete_dos.get_site_spd_dos(site)
        #     plotter.add_dos_dict(spd_dos)

        # NOTE: get_dos_dict may be useful in the future

        plot = plotter.get_plot()
        plot_filename = os.path.join(directory, "density_of_states.png")
        plot.savefig(plot_filename)



# -*- coding: utf-8 -*-

from simmate.workflow_engine import s3task_to_workflow
from simmate.calculators.vasp.tasks.energy import (
    MatProjStaticEnergy as MPStaticEnergyTask,
)
from simmate.calculators.vasp.database.energy import (
    MatProjStaticEnergy as MPStaticEnergyResults,
)

workflow = s3task_to_workflow(
    name="static-energy/matproj",
    module=__name__,
    project_name="Simmate-Energy",
    s3task=MPStaticEnergyTask,
    calculation_table=MPStaticEnergyResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)



# -*- coding: utf-8 -*-

import os

from prefect.utilities.tasks import defaults_from_attrs

from simmate.workflow_engine import S3Task

from simmate.calculators.bader.outputs import ACF
from pymatgen.io.vasp.outputs import Chgcar
from pymatgen.io.vasp import Potcar

# TODO -- I should replace the chgsum.pl script with a simple python function
# that just sums the two files. It might not be as fast but it removes one
# executable file from having to be in the user's path.


class CombineCHGCARsTask(S3Task):
    command = "./chgsum.pl AECCAR0 AECCAR2 > chgsum.out"


class BaderAnalysisTask(S3Task):
    command = "./bader CHGCAR -ref CHGCAR_sum -b weight > bader.out"

    @defaults_from_attrs("dir")
    def setup(self, dir=None):

        # Make sure that there are CHGCAR, AECCAR0, AECCAR2 files from a VASP calc
        files = ["CHGCAR", "AECCAR0", "AECCAR2"]
        filenames = [os.path.join(dir, file) for file in files]
        assert all(os.path.exists(filename) for filename in filenames)

        # Make the CHGCAR_sum file using Bader's helper script
        CombineCHGCARsTask().run(dir=dir)

    @defaults_from_attrs("dir")
    def postprocess(self, dir=None):

        # load the ACF.dat file
        acf_filename = os.path.join(dir, "ACF.dat")
        dataframe, extra_data = ACF(filename=acf_filename)

        # load the electron counts used by VASP from the POTCAR files
        # OPTIMIZE this can be much faster if I have a reference file
        potcar_filename = os.path.join(dir, "POTCAR")
        potcars = Potcar.from_file(potcar_filename)
        nelectron_data = {}
        # the result is a list because there can be multiple element potcars
        # in the file (e.g. for NaCl, POTCAR = POTCAR_Na + POTCAR_Cl)
        for potcar in potcars:
            nelectron_data.update({potcar.element: potcar.nelectrons})

        # grab the structure from the CHGCAR
        # OPTIMIZE I should just grab from the POSCAR or CONTCAR for speed
        chgcar = Chgcar.from_file("CHGCAR")
        structure = chgcar.structure

        # Calculate the oxidation state of each site where it is simply the
        # change in number of electrons associated with it from vasp potcar vs
        # the bader charge I also add the element strings for filtering functionality
        elements = []
        oxi_state_data = []
        for site, site_charge in zip(structure, dataframe.charge.values):
            element_str = site.specie.name
            elements.append(element_str)
            oxi_state = nelectron_data[element_str] - site_charge
            oxi_state_data.append(oxi_state)

        # add the new column to the dataframe
        # !!! There are multiple ways to do this, but I don't know which is best
        # dataframe["oxidation_state"] = pandas.Series(
        #     oxi_state_data, index=dataframe.index)
        dataframe = dataframe.assign(
            oxidation_state=oxi_state_data,
            element=elements,
        )

        # return all of our results
        return dataframe, extra_data



# -*- coding: utf-8 -*-

from django.http import HttpRequest

from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer  # , HyperlinkedModelSerializer

from simmate.website.core_components.filters import DatabaseTableFilter
from simmate.database.base_data_types import DatabaseTable


class SimmateAPIView(GenericAPIView):
    # This class simply adds a method to enable passing extra context to the
    # final Response. This is only done when we are using the HTML format.

    extra_context: dict = {}
    """
    This defines extra context that should be passed to the template when
    using format=html. Note, you can have this as a constant or alternatively
    define a property. The only requirement is that a dictionary is returned.
    """

    def get(self, request: HttpRequest, *args, **kwargs) -> Response:
        if self.viewset_type == "list":
            return self.list_view(request, *args, **kwargs)
        elif self.viewset_type == "retrieve":
            return self.retrieve_view(request, *args, **kwargs)

    def list_view(self, request: HttpRequest, *args, **kwargs) -> Response:

        # This code is modified from the ListModelMixin, where instead of returning
        # a response, we perform additional introspection first.
        # https://github.com/encode/django-rest-framework/blob/master/rest_framework/mixins.py#L33

        # self.format_kwarg --> not sure why this always returns None, so I
        # grab the format from the request instead. If it isn't listed, then
        # I'm using the default which is html.
        self._format_kwarg = request.GET.get("format", "html")

        queryset = self.filter_queryset(self.get_queryset())

        # attempt to paginate the results
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
        else:
            serializer = self.get_serializer(queryset, many=True)

        # If don't have the html format, we follow simple logic from the
        # original ListModelMixin method
        if self._format_kwarg != "html":
            if page is not None:
                return self.get_paginated_response(serializer.data)
            else:
                return Response(serializer.data)

        # otherwise we assume the html format.
        else:
            filterset = self.filterset_class(request.GET)
            data = {
                # "filterset": filterset, # not used at the momemnt
                "filterset_mixins": filterset.get_mixin_names(),
                "form": filterset.form,
                "extra_filters": filterset.get_extra_filters(),
                "calculations": serializer.instance,  # return python objs, not dict
                "ncalculations_matching": queryset.count(),
                "ncalculations_possible": self.get_queryset().count(),
                **self.paginator.get_html_context(),
                **self.extra_context,
            }
            return Response(data)

    def retrieve_view(self, request: HttpRequest, *args, **kwargs) -> Response:

        # self.format_kwarg --> not sure why this always returns None, so I
        # grab the format from the request instead. If it isn't listed, then
        # I'm using the default which is html.
        self._format_kwarg = request.GET.get("format", "html")

        # ---------------------------------------------------
        # This code is from the RetrieveModelMixin, where instead of returning
        # a response, we perform additional introspection first
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        # return Response(serializer.data) <--- removed from original
        # ---------------------------------------------------

        if self._format_kwarg == "html":
            # we want to return a python object, not a serialized dictionary
            # because the object has more attributes and methods attached to it.
            calculation = serializer.instance
            data = {
                "calculation": calculation,
                "table_mixins": calculation.get_mixin_names(),
                "extra_columns": calculation.get_extra_columns(),
                **self.extra_context,
            }
            return Response(data)
        else:
            return Response(serializer.data)


def render_from_table(
    request: HttpRequest,
    template: str,
    context: dict,
    table: DatabaseTable,
    view_type: str,
    request_kwargs: dict = {},
    primary_key_field: str = "id",
    primary_key_url: str = "id",
) -> Response:

    # NOTE: This dynamically creates a serializer and a view EVERY TIME a
    # URL is requested. This means...
    #   1. there is no pre-set api that exists. The exisiting api must be inferred
    #       from lower level workflows and their tables
    #   2. these views will be very inefficient if queried by a script
    # I chose dynamic creation over creating all endpoints on-startup to prevent
    # the `from simmate.database import connect` method from taking too long --
    # as that would require import all workflows on start-up. However, if the
    # (1) api spec or (2) speed of this method ever becomes an issue, I can
    # address these by either...
    #   1. having a utility that prints out the full API spec but isn't called on startup
    #   2. making all APIViews up-front

    # For all tables, we share all the data -- no columns are hidden. Therefore
    # the code for the Serializer is always the same.
    # !!! consider switching to HyperlinkedModelSerializer
    # !!! I may need to better understand this:
    #   https://www.django-rest-framework.org/api-guide/serializers/#how-hyperlinked-views-are-determined
    class NewSerializer(ModelSerializer):
        class Meta:
            model = table
            fields = "__all__"

    NewFilterSet = DatabaseTableFilter.from_table(table)

    # For the source dataset, not all tables have a "created_at" column, but
    # when they do, we want to return results with the most recent additions first
    # by default. Ordering can also be overwritten by passing "ordering=..."
    # to the URL.
    default_ordering_field = (
        "-created_at" if hasattr(table, "created_at") else primary_key_field
    )

    # we also want to preload spacegroup for the structure mixin
    intial_queryset = table.objects.all()
    if hasattr(table, "spacegroup"):
        intial_queryset = intial_queryset.select_related("spacegroup")

    # TODO: consider using the following to dynamically name these classes
    #   NewClass = type(table.__name__, mixins, extra_attributes)
    class NewViewSet(SimmateAPIView):
        # we show the most recent calculations first
        queryset = intial_queryset
        serializer_class = NewSerializer
        template_name = template
        extra_context = context
        filterset_class = NewFilterSet
        viewset_type = view_type  # BUG: view_type = view_type throws an error
        # These two are only used for the view_type = "retrieve"
        lookup_url_kwarg = primary_key_url
        lookup_field = primary_key_field
        # These arguments handle the "ordering" tag passed to the url
        ordering_fields = "__all__"
        ordering = [default_ordering_field]

    # now pull together the html response
    response = NewViewSet.as_view()(request, **request_kwargs)
    return response



# -*- coding: utf-8 -*-

from .files import get_directory, make_archive, archive_old_runs, empty_directory
from .other import get_conda_env, get_doc_from_readme, get_chemical_subsystems

# must be last to prevent circular import
__doc__ = get_doc_from_readme(__file__)



# -*- coding: utf-8 -*-

import os

from simmate.toolkit.diffusion import MigrationHop

from simmate.workflow_engine.workflow import (
    Workflow,
    Parameter,
    ModuleStorage,
)
from simmate.workflow_engine.common_tasks import (
    LoadInputAndRegister,
    parse_multi_command,
)

from simmate.calculators.vasp.workflows.nudged_elastic_band.utilities import (
    get_endpoint_structures,
    get_migration_images_from_endpoints,
)
from simmate.calculators.vasp.workflows.relaxation import (
    neb_endpoint_workflow as relaxation_neb_endpoint_workflow,
)
from simmate.calculators.vasp.workflows.nudged_elastic_band.from_images import (
    workflow as neb_from_images,
)
from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
)

WORKFLOW_NAME = "diffusion/single-path"

# Convert our workflow objects to task objects
relax_endpoint = relaxation_neb_endpoint_workflow.to_workflow_task()
run_neb = neb_from_images.to_workflow_task()

# Extra setup tasks
load_input_and_register = LoadInputAndRegister(
    workflow_name=WORKFLOW_NAME,
    input_obj_name="migration_hop",
)  # TODO: make MigrationHop a calc so we can use calculation_table?


with Workflow(WORKFLOW_NAME) as workflow:

    migration_hop = Parameter("migration_hop")
    directory = Parameter("directory", default=None)
    source = Parameter("source", default=None)

    # This helps link to a higher-level table.
    diffusion_analysis_id = Parameter("diffusion_analysis_id", default=None)
    # TODO: Can the hop id be inferred from the migration_hop or somewhere
    # else in this context? Maybe even load_input_and_register will use
    # prefect id once it's a Calculation?
    migration_hop_id = Parameter("migration_hop_id", default=None)

    # I separate these out because each calculation is a very different scale.
    # For example, you may want to run the bulk relaxation on 10 cores, the
    # supercell on 50, and the NEB on 200. Even though more cores are available,
    # running smaller calculation on more cores could slow down the calc.
    command = Parameter("command", default="vasp_std > vasp.out")
    # command list expects three subcommands:
    #   command_bulk, command_supercell, and command_neb
    subcommands = parse_multi_command(
        command,
        commands_out=["command_supercell", "command_neb"],
    )

    # Load our input and make a base directory for all other workflows to run
    # within for us.
    migration_hop_toolkit, directory_cleaned = load_input_and_register(
        input_obj=migration_hop,
        input_class=MigrationHop,
        source=source,
        directory=directory,
        command=command,
        diffusion_analysis_id=diffusion_analysis_id,
        migration_hop_id=migration_hop_id,
    )

    # get the supercell endpoint structures
    supercell_start, supercell_end = get_endpoint_structures(migration_hop_toolkit)

    # Relax the starting supercell structure
    run_id_00 = relax_endpoint(
        structure=supercell_start,
        command=subcommands["command_supercell"],
        directory=directory_cleaned + os.path.sep + "endpoint_relaxation_start",
    )

    # Relax the ending supercell structure
    run_id_01 = relax_endpoint(
        structure=supercell_end,
        command=subcommands["command_supercell"],
        directory=directory_cleaned + os.path.sep + "endpoint_relaxation_end",
    )

    images = get_migration_images_from_endpoints(
        supercell_start={
            "calculation_table": "NEBEndpointRelaxation",
            "directory": run_id_00["directory"],
            "structure_field": "structure_final",
        },
        supercell_end={
            "calculation_table": "NEBEndpointRelaxation",
            "directory": run_id_01["directory"],
            "structure_field": "structure_final",
        },
    )

    # Run NEB on this set of images
    run_id_02 = run_neb(
        migration_images=images,
        command=subcommands["command_neb"],
        source=source,
        directory=directory_cleaned,
        diffusion_analysis_id=diffusion_analysis_id,
        migration_hop_id=migration_hop_id,
    )


workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-Diffusion"
# workflow.calculation_table = MITDiffusionAnalysis  # not implemented yet
# workflow.register_kwargs = ["prefect_flow_run_id"]
workflow.result_table = MITDiffusionAnalysis
workflow.s3tasks = [
    relaxation_neb_endpoint_workflow.s3task,
    neb_from_images.s3task,
]

workflow.__doc__ = """
    Runs a full diffusion analysis on a single migration hop using NEB.
    
    Typically, this workflow is not ran directly -- but instead, you would
    use diffusion/all-paths which submits a series of this workflow for you.
    
    For a given migration path, supercell structures are generated for
    the start and end points of the migration, where vacancy-based diffsion
    is used. These supercells are relaxed using the relaxation/neb-endpoint 
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images
    
    Currently, this workflow will not run through the command-line, as we
    have not implemented a file format for MigrationHop's yet.
"""



# -*- coding: utf-8 -*-

from django.shortcuts import render


def home(request):
    context = {}
    template = "website/home.html"  # !!! Doesn't exist at the moment
    return render(request, template, context)



# -*- coding: utf-8 -*-

from prefect import Client


def build():

    # grab the Prefect client
    client = Client()

    # BUG: for now we assume we are in the correct tenant. If not, we'd need to run:
    #   client.login_to_tenant(tenant_slug='a-tenant-slug')

    # grab all of the workflows that need to be registered
    from simmate.workflows.utilities import get_list_of_all_workflows, get_workflow

    # Load all workflows
    workflow_names = get_list_of_all_workflows()
    workflows = [get_workflow(name) for name in workflow_names]

    # Iterate through and grab all the unique Project names. The brackets here
    # make this return as a set
    project_names = {workflow.project_name for workflow in workflows}

    # build each of the projects
    for project_name in project_names:
        project_id = client.create_project(project_name=project_name)
    # register the workflows with the proper projects
    for workflow in workflows:
        workflow_id = workflow.register(
            project_name=workflow.project_name,
            set_schedule_active=False,
        )
    # TODO: will I need to store the project or workflow IDs? Maybe inside
    # the /simmate/prefect folder? Or maybe even the /.prefect folder?
    # I could also return a list of what's been done.


def delete():

    # grab the Prefect client
    client = Client()

    # query for a list of projects that start with "Simmate-"
    # BUG: I delete all projects for now. I need to update this query to just
    # Simmate projects
    query = "query {project {name}}"
    result = client.graphql(query)
    project_names = [p["name"] for p in result["data"]["project"]]

    # delete all of the Simmate projects
    # BUG: if there's too much data in any of these projects, then the command
    # may time-out. In that case, I may need to delete some flow runs manually
    # before I can delete the project.
    for project_name in project_names:
        try:
            is_successful = client.delete_project(project_name=project_name)
        except ValueError:
            # an error will be raised if the project isn't found -- so if it doesn't
            # exist, it's already deleted. We can just move on in this case.
            is_successful = True
            pass

    # In the scenario we already deleted everything, just say we were successful
    if not project_names:
        is_successful = True

    # BUG advisory
    print(
        "WARNING: Deleting projects in Prefect Cloud is a 'lazy' mutation. This means the "
        "project is only hidden initially and the deletion could take a few minutes "
        "to complete in Prefect's backend. Therefore, you'll need to wait some "
        "time before trying to recreate your projects with build(). 1 minute should "
        "be plenty of time, but the Prefect team says this can take as much as 30+ min."
    )

    return is_successful


# BUG: calling delete and then build back-to-back causes errors because delete
# is a lazy operation. Therefore, users need to "sleep here for some time"
# def reset():
#     # reseting just involves deleting all projects and then recreating them
#     delete()
#     build()



# -*- coding: utf-8 -*-

from .mit import workflow as mit_workflow
from .materials_project import workflow as matproj_workflow
from .quality_00 import workflow as quality00_workflow
from .quality_01 import workflow as quality01_workflow
from .quality_02 import workflow as quality02_workflow
from .quality_03 import workflow as quality03_workflow
from .quality_04 import workflow as quality04_workflow
from .staged import workflow as staged_workflow
from .neb_endpoint import workflow as neb_endpoint_workflow
